

! -*-f90-*
! Originator:cbrindle
! Adapted:   Jorgo Bakker (jbakker@rssd.esa.int)
! Info:
!   Interactive program for selecting sources from an image
!   for further analysis
! Note:
!  * Should always call "omsource_init()" first, and "omsource_finish()" last.
!  * In comments, I usually talk about methods. A method may be either a
!    function OR subroutine within a module
!  * Any variable starting with 'my' is global to this module
   

module omsource_old
   use dal
   use cal
   use sasdal
   use oal
   use param
   use sla
   use errorHandling
   
   use frametime_mod, only : getFrameTime ! mssllib
   use getFilterID_mod
   use omutility_mod      ! In mssllib various fits access routines
   use omconstants_mod    ! In mssllib defines various constants
   use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
   use omparameters_mod 
   use omphotometry_mod   ! In mssllib for doing aperture photometry
   use ommagnitude_mod    ! In mssllib for computing instrumental magnitude 
   use omsrctable_mod     ! In mssllib for creating the source-list table
   use omsourcecelestial_mod
   use omprogress_mod
   use omsourceregion_mod
   use omsourceparameter_mod      
   use omsourceflags_mod   
   use omskybackgroundimage_mod
   use ominterpolate_mod
   use leastsquares_mod
   use omminim_mod
   implicit none


   private ! all, but:

   public :: myImage
   public :: myBackgroundImage
   public :: myBackgroundSigmaImage
   
   public :: myLevelImage
   public :: myExposureImage 
   public :: OmSourceRegion      ! type definition
   public :: ObsDetails           ! NB
   public :: ImageWindow          !

   public  :: omsource_init       ! subroutine ...()
   public  :: omsource_finish     ! subroutine ...()
   public  :: omsource_loadsources  ! subroutine ... ()
   public  :: omsource_delete     ! subroutine ...()
   public  :: omsource_deleteList 

   public  :: omsource_replace  ! subroutine ...()
   public  :: omsource_recompute  ! subroutine ...()
   public  :: omsourceold_recomputephotometry  ! subroutine ...()


   public  :: omsource_add        ! subroutine ...()
   public  :: omsource_add1        ! subroutine ...()
   public  :: omsource_add2        ! subroutine ...()

   public  :: omsource_computeparameters
   public  :: omsource_computesourcepositions


   public  :: omsource_photometry
   public :: omsource_getbackgroundmode

   public  :: omsource_addBackgroundSources

   public  :: omsource_getnewsources      ! subroutine ...()
   public :: setTheSourceFlags
 
   interface construct
      module procedure construct_od
   end interface

   interface destruct
      module procedure destruct_od
   end interface
 
   interface construct
      module procedure construct_iw
   end interface

   interface destruct
      module procedure destruct_iw
   end interface


   interface construct
      module procedure construct_celestial
   end interface

   interface destruct
      module procedure destruct_celestial
   end interface



    !***************************************************************************
    ! The following is for FAST data, if using a separate image for background
    !***************************************************************************
    integer, private :: x0Image, y0Image, imageBinx, imageBiny 
    integer, private :: x0Background, y0Background, backgroundBinx, backgroundBinY
    logical, private :: fastBackground = .false.


    !*******************************************************************************
    !
    ! The following are for PSF photometry
    !
    !*******************************************************************************
    real(single), private, pointer :: mypsfImage(:, :) => null()
    integer(int8), private, parameter :: maxNoSources = 5
    integer, private :: imin, imax, jmin, jmax
    real(single), private , pointer, dimension(:, :) :: xdata => null()
    real(single), private,  pointer, dimension(:) :: ydata => null(),  &
           weight => null()

    integer(int16), private :: l0=0

    real(single), private :: psfxCentre, psfYcentre, sumPsfImage
    real(single), private :: meanBackground = 0.0

    real(double), private :: bl(10), bu(10)
    integer(int8), private, parameter :: maxShift = 4

    integer, private :: psfNx, psfny
    integer(int8), private :: numberToFit, weightingMethod = 1



real(single), private :: exposureConstant = 0.0
real(double), private :: meanExposure = 0.0_double
integer, private :: numExposurePixels = 0

interface extendedSourcePhotometry
    module procedure extendedSourcePhotometry1, extendedSourcePhotometry2 
end interface extendedSourcePhotometry




contains
!********************************************************
!
!*******************************************************
subroutine setweightingmethod( method )
    integer, intent(in) :: method

    weightingMethod = method

end subroutine setweightingmethod



!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine omsource_init(sources, backgroundSources, &
                         addedSources, addedBackgroundSources, &
                         photometricParameters, nx, ny)

    type(SourceParameters),  pointer    :: sources(:)
    type(SourceParameters),  pointer    :: backgroundSources(:)
    type(SourceParameters),  pointer    :: addedSources(:)
    type(SourceParameters),  pointer    :: addedBackgroundSources(:)


    type(PhotometryParameters), intent(out) :: photometricParameters
    integer, intent(inout) :: nx, ny      
    type(OmProgressT)        :: reporter    ! ... of progress to GUI
    character(*) , parameter :: routine="omsource_old.omsource_init"
    real(single) :: sigma 
     
    call construct(reporter, label = "Initialising...", steps = 20)
    call setProgress(reporter,1)


    nullify(sources)       ! proper state
    nullify(backgroundSources)
    nullify(addedSources)
    nullify(addedBackgroundSources)

    ! --- initialise globals in this module


    myDeadFraction       = -1.
    myFrameTime          = -1.
    myDetectionRate      = -1.
    myDetectionMagnitude = -1.

    nullify(myImage, myBackgroundimage, mybackgroundsigmaimage)
    nullify(myQuality, myLevelImage, myExposureImage)

   
    !**********************************************
    ! Extract the image from the file
    !**********************************************
    myImageSet = dataSet(stringParameter("imageset"), read)


    !********************************************************************
    ! load the image parameters (WINDOWX, etc) from the image fits file...
    !********************************************************************
    call construct(myImageWindow, set = myImageSet)





    
    ! set the frame time etc...


    myFrameTime = myImageWindow%frameTime
    myDeadFraction = myImageWindow%deadFraction
    frameTime = myFrameTime
    deadFrac = myDeadFraction

    call setProgress(reporter,4)


    ! Get the observation details for plotting
    call construct(myObsDetails, set = myImageSet)

    !<TODO: missing DATATYPE in dataset, otherwise we could do this:
    !call CAL_setState(myImageSet)


    call CAL_setState(OM, filterId = filterIdToInt(trim(myObsDetails%filter)))
    call setSignificanceRadius( trim(myObsDetails%filter) ) 


    !call setProgress(reporter,5)
    !TODO>

    ! create celestial information; also sets state of CAL
    call construct(myCelestial, set=myImageSet)
    call setProgress(reporter, 6)

    ! now init radii once, and once only, and bin them now
    myApertureRadius = 12.0                ! NEED THIS IN UNBINND PIXELS

    myCalRadius = CAL_omGetApertureRadius() ! NEED THIS IN UNBINNED PIXELS




    if (myCalRadius .gt. myApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        myScaleCounts = .true.
        myScaleCountsRatio = (myCalRadius / myApertureRadius) ** 2 
    end if

    photometricParameters%apertureRadius =  myApertureRadius / real(myImageWindow%bx)
    
    photometricParameters%calApertureRadius =  myCalRadius /  real(myImageWindow%bx)
    photometricParameters%binX = myImageWindow%bx
    imageBinx =  myImageWindow%bx
    imageBiny =  myImageWindow%by

    myBgrRadiusIn    = photometricParameters%calApertureRadius + 2.0

    myBgrRadiusOut   = sqrt(3.0 *   photometricParameters%calApertureRadius * &
          photometricParameters%calApertureRadius +  myBgrRadiusIn  * myBgrRadiusIn)
  

    !************************************* 
    ! Get the Plate scale (arcsecs/pixel)
    !*************************************
    call CAL_omGetPlateScale(0.55_double, myImageWindow%plateScale)


     
    call setProgress(reporter, 7)

    ! Get the RA and DEC of the image centre
!      call setLabel(reporter,"Computing sky coordinates")
    call getCelestialCentre(myCelestial, myImageWindow%RA, myImageWindow%Dec)
    call setProgress(reporter, 8)


    !*******************************************************
    ! Get the image and, if present, the exposure image
    !********************************************************
    call loadTheImage( myImageSet )




    ! Compute the background mode

    nx = size(myImage, 1)
    ny = size(myImage, 2)
 
    
    myBgrMode = backgroundMode(myImage, int(nx, int16), int(ny, int16), 2.0, sigma) 



    !****************************************************************
    ! See if a background image has been specified
    !
    !*****************************************************************

    if ( len( trim( stringParameter("backgroundimageset"))) > 2) then
        call loadBackgroundImage
    else
        myBackgroundImageIsLoaded = .true.
      
        fastBackground = .false.
    end if 



    nx = nx - 1
    ny = ny - 1
   
    myLevelImage => allocate2dInt16(0, nx, 0, ny, "myLevelImage")
   
    call setProgress(reporter, 9)

    ! Get the quality of the pixels
    call createQualityArray(myImageSet, myQuality)
    call setProgress(reporter, 10)

     
    call destruct(reporter)
    ! --- finished general information


    ! Load the old sourcelist file if one has been given on the
    ! command line
    
    call loadSources(sources, stringParameter("oldsrclistset"))

    mySourcesAreLoaded = associated(sources)


    !********************************************************
    ! For initialising psf photometry
    !*******************************************************
    call initialisePSFPhotometry()



!write(*, *) "omsource_old line 381"

end subroutine omsource_init

!*******************************************************************************
! Info: Loads the sources from the source-list file produced by omdetect & ommag
!       The 'sources' argument is populated with the detected sources.
!*******************************************************************************
subroutine omsource_loadsources(sources, sourceListFile)
    type(SourceParameters),  pointer    :: sources(:)
    
    character(256), intent(in) :: sourceListFile;
 
 
    call loadSources(sources, sourceListFile)
     
     
    ! Compute the RA and Declination of each source

!      call setLabel(reporter,"Celestial properties")      
    call setCelestial(myCelestial, sources)

end subroutine omsource_loadsources

!*************************************************************
! omsource_recompute
!
!
! Info: recomputes the properties for a source=sources(id)
!***************************************************************
subroutine omsource_recompute(sources, id )
    type(SourceParameters),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    integer(int16) :: nx, ny
    type(SourceParameters), pointer :: s
    type(SourceParameters), pointer  :: tmp(:) => null()
    integer(int32)                  :: bin(2)
    character(*) , parameter :: routine="omsource_old.omsource_recompute"
    real(single) :: corrBackgroundCountRateErr = 0.0
    real(double) :: ra, dec   

!write(*, *) "omsource-old - omsource-recompute line 404"
    s=>sources(id)
    s%exposure = myobsDetails%exposure

    ra = s%rightAscension
    dec = s%declination

    nx = size(myImage, 1)
    ny = size(myImage, 2)

    bin= (/myImageWindow%bx, myImageWindow%by/)
 

    ! Do aperture photometry on  source

    allocate(tmp( 1 : 1))
    tmp(1) = s



!write(*, *) "omsource recompute inner backgroundradius=", s%innerbckradius
!write(*, *) "omsource recompute outer backgroundradius=", s%outerbckradius
!write(*, *) "line 348 imin=", s%imin



    if ( s%extensionflag .eq. 0) then



!write(*, *) "omsource line 352 apertureradius=", s%apertureRadius
!write(*, *) "background value=", s%rawBackgroundCountRate

!write(*, *) "exposure=", s%exposure, " x=", s%xpos, " y=", s%ypos
!write(*, *) "omsource line 365 input apertureradius=", s%apertureRadius
!write(*, *) "omsource line 365 tmp input apertureradius=", tmp(1)%apertureRadius
!write(*, *) "coinclossradius=", coinclossradius
!write(*, *) "Input backgroundrawcountrate=", tmp(1)%rawbackgroundcountrate
!write(*, *) "Input exposure=", tmp(1)%exposure


        call AperturePhotometry(int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
            myImage, mylevelImage,                                                          &
            myQuality, myapertureRadius / real(binX),                                       &  
            1_int16, tmp)                  


!write(*, *) "output backgroundrawcountrate=", tmp(1)%rawbackgroundcountrate
!write(*, *) "output rawcountrate=", tmp(1)%rawcountrate


        s = tmp(1)

        !************************************************************************
        ! NB after calling aperture photometry the aperture radius will have been
        ! multiplied by the binning factor binx
        !************************************************************************


  

!write(*, *) "omsource line 371 apertureradius=", s%apertureRadius


    else

!write(*, *) "omsource-old - omsource-recompute line 470 imin=", s%imin


        select case (s%imin)


            case (-1)
 
!write(*, *) "omsourceold line 478 calling  OmsourceExtendedSourceCounts1- interplaotion using background annulus pixels"
                !************************************************************************************
                ! Intepolation using background annulus pixels to remove background
                !***********************************************************************************
   
                call OmsourceExtendedSourceCounts1( myImage, tmp(1), myApertureRadius / real(binx), &
                  s%apertureRadius * s%axialRatio, s%apertureRadius, s%theta)



            case(-2)

!write(*, *) "Line 490 calling  OmsourceExtendedSourceCounts2 using annulus pixels to compute background"

                !************************************************************************************
                ! Background from background annulus pixels
                !***********************************************************************************
                call OmsourceExtendedSourceCounts2( myImage, tmp(1), myapertureRadius / real(binx), &
                  s%apertureRadius * s%axialRatio, s%apertureRadius, s%theta)
 
            case(-3)

!write(*, *) "Line 500 calling  OmsourceExtendedSourceCounts2 using fixed background"

                !************************************************************************************
                ! Background from fixed background counts/pixel
                !***********************************************************************************
                call OmsourceExtendedSourceCounts3( myImage, tmp(1), myapertureRadius / real(binx), &
                  s%apertureRadius * s%axialRatio, s%apertureRadius, s%theta)




           end select

     
    end if 
    s = tmp(1)


    !*******************************************************************************
    !
    !
    !*******************************************************************************
    if( associated(myExposureImage) .and. s%extensionFlag == 1 ) then
        call scaleRateErrors( s )
    end if 


!   write(*, *) "line 434 apertureradius=", myApertureRadius, " source aperture radius=", s%apertureRadius


  
    s%semiberr = s%rightascension
    s%semiaerr = s%declination
    s%rightAscension = ra
    s%declination = dec

!   write(*, *) "omsource_old line 421 BEFORE calling compute magnitude - s%corrcountrate=",s%corrCountRate 
!  write(*, *) "s%rawBackgroundCountRate=",  s%rawBackgroundCountRate
!   write(*, *) "s%rawCountRate=",  s%rawCountRate

!   write(*, *) "s%corrBackgroundCountRate=",  s%corrBackgroundCountRate
!   write(*, *) " frametime=", myFrameTime, " deadfraction=", myDeadFraction
!   write(*, *) "line 446 apertureradius=", myApertureRadius, " source aperture radius=", s%apertureRadius
!   write(*, *) " tdscorr=",s%tdsCorr
    call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr,                   &
                          s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                          s%xPos, s%yPos,                                        &
                          myScaleCounts, s%extensionFlag,                        &
                          myCalRadius, myApertureRadius,                         &
                          myFrameTime, myDeadFraction,                           &
                          s%magnitude, s%magnitudeErr, s%psf2Correction,         &
                          myobsDetails%filter, myObsDetails%mjd, s%tdsCorr)

    s%apertureRadius = s%apertureRadius / real(binX)
 

!write(*, *) "Line 435 computed magnitude=", s%magnitude, " apertureradois=", s%apertureRadius, &
!" tdscorr=", s%tdscorr, " mjd=", myObsDetails%mjd

    sources(id) = s

!write(*, *) "461 sourceid apertureradois=", sources(id)%apertureRadius

    deallocate(tmp)
    nullify(tmp)
end subroutine omsource_recompute
!*************************************************************
! omsource_delete
!
!
! Info: deletes source=sources(id) from the source file; then all sources are
!       renumbered to make the IDs consitent.
!************************************************************/
subroutine omsource_delete(sources, id)
    type(SourceParameters),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    type(SourceParameters),  pointer :: v(:)
    integer                           :: l, h  ! lbound/ubound
    integer                           :: idx  ! iterator
    integer                           :: off  ! 

    ! assume that sources(id) exists
    l = maxval(lbound(sources))
    h = maxval(ubound(sources))
    off = 0

    call new(v, low = l, high = h - 1)
    do idx = l, h
        if (idx .eq. id) then
            off = -1
            cycle
        end if
        v(idx + off) = sources(idx)
    end do

    call delete(sources)
    sources => v
    nullify(v)

end subroutine omsource_delete
!***********************************************
! Deletes the whole source list
!**********************************************
subroutine omsource_deleteList(sources)

    type(SourceParameters),  pointer    :: sources(:)
!write(*, *) "omsource_old line 404 deletelist"
    if (.not. associated(sources)) return
    call delete(sources)   
    nullify(sources)


    call new(sources, low = 1, high = 0)

end subroutine omsource_deleteList

!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add(sources, regions, radius, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    real(single), intent(in) :: radius
    type(SourceParameters),  pointer    :: newSources(:)

    ! locals
    integer                         :: i
    integer,                pointer :: idx(:)
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

    
    call omsource_deleteList(newSources)
       
!write(*, *) "omsource-old line 717 -omsource_add- exposure=", myObsDetails%exposure
   
    ! --- check new regions against current sources
    nullify(added)
    if (associated(sources)) then
        idx => getNewRegions(regions, sources, real(16 / myImageWindow%bx))
        if (associated(idx)) then
            call new(added, low = 1, high = size(idx))

            do i = 1, size(idx)
                added(i)%xPos=   regions(idx(i))%x
                added(i)%yPos=   regions(idx(i))%y
                added(i)%exposure= myObsDetails%exposure
                added(i)%apertureRadius = radius
            end do

         
              deallocate(idx)
    
       end if
       nullify(idx)
    else
        call new(added,low = 1, high = size(regions))

        do i = 1, size(regions)
            added(i)%xPos=   regions(i)%x
            added(i)%yPos=   regions(i)%y
            added(i)%exposure= myObsDetails%exposure
     
  

        end do
      
    end if
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then


        call ComputeParameters(added)
        call new(newSources, low = 1, high = size(added))
        newSources = added
      

    
    end if
    call setProgress(reporter, 3)



    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if


    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_add


!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add1(sources, regions, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    type(SourceParameters),  pointer    :: newSources(:)

    ! locals
    integer                         :: i
  
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

    
    call omsource_deleteList(newSources)
       
  
    ! --- check new regions against current sources
    nullify(added)
 
    call new(added, low = 1, high = size(regions))
!write(*, *) "omsource-old line 553"
    do i = 1, size(regions)
        added(i)%xPos=   regions(i)%x
        added(i)%yPos=   regions(i)%y
        added(i)%exposure= myObsDetails%exposure
       
!      write(*, *) i, " x=", added(i)%xPos, " y=", added(i)%yPos 
    end do
      
   
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then

        ! Compute the source positions, etc
     

        call omsource_ComputeSourcePositions(added)
   
    
        ! Compute RAs and decs
        call setCelestial(myCelestial, added)
        call new(newSources, low = 1, high = size(added))
        newSources = added
        

    
    end if

    call setProgress(reporter, 3)



    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if

   
   
    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_add1


!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add2(sources, regions, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    type(SourceParameters),  pointer    :: newSources(:)

    ! locals
    integer                         :: i
  
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

    
    call omsource_deleteList(newSources)
       
  
    ! --- check new regions against current sources
    nullify(added)
 
    call new(added, low = 1, high = size(regions))

    do i = 1, size(regions)
        added(i)%xPos=   regions(i)%x
        added(i)%yPos=   regions(i)%y
        added(i)%exposure= myObsDetails%exposure
        added(i)%semia =  real(4 / myImageWindow%bx)
        added(i)%semib=   added(i)%semia
        added(i)%theta = 0.0
    
      
    end do
      
   
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then

        ! Compute the source positions, etc
  

   
         
!        call omsource_ComputeSourcePositions(added)
   
    
        ! Compute RAs and decs
        call setCelestial(myCelestial, added)
        call new(newSources, low = 1, high = size(added))
        newSources = added
        

    
    end if

    call setProgress(reporter, 3)



    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if
   

    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_add2








!**************************************************************
! Info: add regions to the background source list
! ignoring "sources"
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_addBackgroundSources(sources, &
        backgroundSources, regions, apertureRadius)
    type(SourceParameters),  pointer    :: sources(:)
    type(SourceParameters),  pointer    :: backgroundSources(:)

    type(OmSourceRegion),     intent(inout) :: regions(:)
    real(single), intent(in) :: apertureRadius
    ! locals
    integer                         :: i, j, k, k1
    integer,                pointer :: idx(:)
    real(single)  :: radius
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = size(regions))
    call setProgress(reporter,1)


    ! --- check new regions against current sources
    nullify(added)
    if (associated(backgroundSources)) then
        idx => getNewBackgroundRegions(regions, sources, backgroundSources, real(16 / myImageWindow%bx))


        if (associated(idx)) then
            call new(added, low = 1, high = size(idx))
            do i = 1, size(idx)
                call setProgress(reporter, i)

                added(i)%xPos = regions(idx(i))%x
                added(i)%yPos = regions(idx(i))%y
                added(i)%exposure = myObsDetails%exposure   
                
                added(i)%apertureRadius = apertureRadius;
                added(i)%semia = apertureRadius;
                added(i)%semib = apertureRadius;
                added(i)%theta = 0.0;
        

            end do
            deallocate(idx)
        end if
        nullify(idx)
    else
        iLoop : do i = 1, size(regions)
            k = 0            
            do j = 1, size(regions)
                if (i == j) cycle
                radius = sqrt((regions(i)%x - regions(j)%x) ** 2 + (regions(i)%y - regions(j)%y) ** 2)
                if (radius < 1.0) cycle iLoop
                  
             end do 
             k = k + 1
             regions(k) = regions(i)
        end do iLoop

        if (associated(sources)) then
            k1 = 0
            iLoop1 : do i = 1, k 
                do j = 1, size(sources)
                    radius = sqrt((regions(i)%x - sources(j)%xpos) ** 2 + (regions(i)%y - sources(j)%xpos) ** 2)
                    if (radius < 1.0) cycle iLoop1
                end do
                k1 = k1 + 1
                regions(k1) = regions(i)    
            end do iLoop1  
            k = k1
        end if
        
        if (k > 0) then
            call new(added, low = 1, high = k)
            do i = 1, k
            
                added(i)%xPos = regions(i)%x
                added(i)%yPos = regions(i)%y
                added(i)%exposure = myObsDetails%exposure
 
                added(i)%apertureRadius = apertureRadius;
                added(i)%semia = apertureRadius;
                added(i)%semib = apertureRadius;
                added(i)%theta = 0.0;
     


             end do
         end if
    end if
  

    ! compute properties for new sources
    if (associated(added)) then


!        call ComputeParameters(myImage, added)
        
    end if
    call setProgress(reporter, 3)

    ! merge
    if (associated(backgroundSources) .and. associated(added)) then
        
        tmp=>backgroundSources

        call new(backgroundSources, low = 1, high = size(tmp) + size(added))
        backgroundSources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            backgroundSources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) backgroundSources=>added
        nullify(added)
    end if

    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_addBackgroundSources

!**************************************************************
! Info: gets the new regions
! 
!
!**************************************************************
subroutine omsource_getnewsources(sources, regions, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    type(SourceParameters),  pointer    :: newSources(:)


    ! locals
    integer                         :: i
    integer,                pointer :: idx(:)
   
  
   

    ! --- check new regions against current sources
    if (associated(newSources)) call delete(newSources)

    nullify(newSources)

    if (associated(sources)) then
    

        idx => getNewRegions(regions, sources, real(16 / myImageWindow%bx))
        if (associated(idx)) then
            call new(newSources, low = 1, high = size(idx))
            do i = 1, size(idx)
                newSources(i)%xPos=   regions(idx(i))%x
                newSources(i)%yPos=   regions(idx(i))%y

                
                newSources(i)%exposure= myObsDetails%exposure
                newSources(i)%apertureRadius = 0.0 ! myApertureRadius;
                newSources(i)%semia = myApertureRadius;
                newSources(i)%semib = myApertureRadius;
                newSources(i)%theta = 0.0;
              
            end do

  


            deallocate(idx)
        end if
        nullify(idx)
    else
        call new(newSources, low = 1, high = size(regions))
        do i = 1, size(regions)
            newSources(i)%xPos =   regions(i)%x
            newSources(i)%yPos =   regions(i)%y
            newSources(i)%exposure = myObsDetails%exposure
            newSources(i)%semia = myApertureRadius;
            newSources(i)%semib = myApertureRadius;
            newSources(i)%theta = 0.0;
           
        end do
    end if



end subroutine omsource_getnewsources

!******************************************************
!
!*****************************************************
subroutine omsource_replace(id, source, sources)
    integer, intent(in) :: id
    type(SourceParameters), intent(in) :: source
    type(SourceParameters),  pointer    :: sources(:)
      
    sources(id) = source

end subroutine omsource_replace 
!******************************************************
! Info: saves sources, and frees memory
! used globals:
!   myCelestial
!   myImage
!   myImageSet
!   myImagyeWindow
!   myObsDetails
!   myQuality
!******************************************************
subroutine omsource_finish(sources, oldSources, backgroundSources, &
                           addedSources, addedBackgroundSources    )
    type(SourceParameters),  pointer    :: sources(:)
    type(SourceParameters),  pointer    :: oldSources(:)

    type(SourceParameters),  pointer    :: backgroundSources(:)
    type(SourceParameters),  pointer    :: addedSources(:)
    type(SourceParameters),  pointer    :: addedBackgroundSources(:)


    ! Add the sources to a table in the new source-list file
    call omsource_save(sources,&
                       stringParameter("newsrclistset"),&
                       myImageSet,&
                       myImageWindow)
    ! Close the image set, as we don't need it anymore.
    call release(myImageSet)

    ! Free memory for this module globals
    call destruct(myCelestial)
    call destruct(myImageWindow)
    call destruct(myObsDetails)

    if (associated(myImage))   deallocate(myImage)
    nullify(myImage)

    if (associated(myExposureImage))   deallocate(myExposureImage)
    nullify(myExposureImage)



    if (myBackgroundImageIsLoaded ) then
        if (associated(myBackgroundImage))   deallocate(myBackgroundImage)

    end if
    nullify(myBackgroundImage)

    if (associated(myBackgroundSigmaImage))   deallocate(myBackgroundSigmaImage)

    nullify(myBackgroundSigmaImage)

    if (associated(mylevelImage))   deallocate(mylevelImage)
    nullify(myLevelImage)

    if (associated(myQuality)) deallocate(myQuality)
    nullify(myQuality)

    ! And the sources themselves
    call delete(sources)
    call delete(oldSources)
    
    call delete(backgroundSources)
    call delete(addedSources)
    call delete(addedBackgroundSources)
  


end subroutine omsource_finish



!*******************************************************************************
!--------------- helper functions private to this module -------------------
!
!
! Info: Compares the regions against the existing sources, and returns a 
!       vector containing indices to unique regions
! Note: * looks like the old MatchPoints subroutine, but cleaned-up
!       * returns a pointer, that must be destroyed by the caller
!********************************************************************************
function getNewRegions(r, s, toleranceIn) result(idx)
    type(OmSourceRegion),   intent(in)    :: r(:)
    type(SourceParameters), intent(in)    :: s(:)
    integer,                pointer       :: idx(:)
    real(single), optional, intent(in)    :: toleranceIn

    ! locals
    real(single)                          :: radius, tolerance
    integer                               :: i, j, nToKeep
    logical                               :: keep(size(r))
 

    ! init
    nullify(idx) ! always be on the safe side ;-)
    if (present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 5.0
    end if
    nToKeep = size(keep)

  

   ! populate the 'keep' and count the unique points
    do i = 1, size(r)
        keep(i) = .true.

        do j = 1, size(s)
            radius = sqrt((s(j)%xPos - r(i)%x) ** 2 +  (s(j)%yPos - r(i)%y)** 2)
            if (radius .le. tolerance) then ! overlap
               keep(i) = .false.
               nToKeep = nToKeep - 1

               exit ! inner loop
            end if
        end do
    end do
    if (nToKeep .eq. 0) return

    ! now we create an index of all regions that are new
    allocate(idx(nToKeep))
!    call assertOn(status/=0,"omsource_old.getNewRegions.idx")

    j = 0
    do i = 1, size(r)
        if (.not. keep(i)) cycle
        j = j + 1
        idx(j) = i
    end do
end function getNewRegions






!*******************************************************************************
!--------------- helper functions private to this module -------------------
!
!
! Info: Compares the regions against the existing sources, and returns a 
!       vector containing indices to unique regions
! Note: * looks like the old MatchPoints subroutine, but cleaned-up
!       * returns a pointer, that must be destroyed by the caller
!********************************************************************************
function getNewBackgroundRegions(region, source, background, toleranceIn) result(idx)
    type(OmSourceRegion), intent(in)      :: region(:)
    type(SourceParameters), pointer    :: source(:)
    type(SourceParameters), pointer    :: background(:)

    integer,                pointer       :: idx(:)
    real(single), optional, intent(in)    :: toleranceIn

    ! locals
    real(single)                          :: radius, tolerance
    integer                               :: i, j, nToKeep
    logical                               :: keep(size(region))
    logical, pointer                      :: identicalSource(:) => null()
    integer                               :: status


    ! init
    nullify(idx) ! always be on the safe side ;-)
    if (present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 5.0
    end if
    nToKeep = size(keep)



    ! populate the 'keep' and count the unique points
    if (associated(source)) then

    
        allocate(identicalSource(1 : size(source)))
        identicalSource = .false.
        ! Check for 2 identical sources




        do i = 1, size(source)
            if (identicalSource(i)) cycle
            
            do j = 1, size(source)
                
                if (i .ne. j .and. .not. identicalSource(j)) then
                    radius = sqrt((source(j)%xPos - source(i)%xPos) ** 2 +  (source(j)%yPos - source(i)%yPos)** 2)

                    if (radius .le. tolerance) then ! overlap
                        identicalSource(j) = .true.  ! Just keep source i
                       
                    end if
                end if
            end do
        end do
    
  

        do i = 1, size(region)
                
            keep(i) = .true.

            do j = 1, size(source)
                if (identicalSource(j)) then
 
                    keep(i) = .false.
                    nToKeep = nToKeep - 1

                    exit ! inner loop
                end if 
                radius = sqrt((source(j)%xPos - region(i)%x) ** 2 +  (source(j)%yPos - region(i)%y)** 2)
                if (radius .le. tolerance) then ! overlap
                    keep(i) = .false.
                    nToKeep = nToKeep - 1

                    exit ! inner loop
                end if
            end do
        end do
        deallocate(identicalSource)
    end if

    

    allocate(identicalSource(1 : size(background)))
    identicalSource = .false.
    ! Check for 2 identical sources
    do i = 1, size(background)
        if (identicalSource(i)) cycle
            
        do j = 1, size(background)
                
            if (i .ne. j .and. .not. identicalSource(j)) then
                radius = sqrt((background(j)%xPos - background(i)%xPos) ** 2 + &
                        (background(j)%yPos - background(i)%yPos)** 2)

                if (radius .le. tolerance) then ! overlap
                    identicalSource(j) = .true.  ! Just keep source i
                       
                end if
            end if
        end do
    end do


    do i = 1, size(region)

        if (.not. keep(i)) cycle

        do j = 1, size(background)


            if (identicalSource(j)) then
 
                keep(i) = .false.
                nToKeep = nToKeep - 1

                exit ! inner loop
            end if 
         

            radius = sqrt((background(j)%xPos - region(i)%x) ** 2 +  (background(j)%yPos - region(i)%y)** 2)
            if (radius .le. tolerance) then ! overlap
               keep(i) = .false.
               nToKeep = nToKeep - 1

               exit ! inner loop
            end if
        end do
    end do

    deallocate(identicalSource)

    if (nToKeep .eq. 0) return

    ! now we create an index of all regions that are new
    allocate(idx(nToKeep), stat = status)
!    call assertOn(status/=0,"omsource_old.getNewRegions.idx")

    j = 0
    do i = 1, size(region)
        if (.not. keep(i)) cycle
        j = j + 1
        idx(j) = i
    end do


end function getNewBackgroundRegions




!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(sources, setName)

    type(SourceParameters), pointer    :: sources(:)
    character(len=*), intent(in) :: setName
   
  
    ! locals
    integer(int32)        :: nSources
  
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH
    real(single)     :: innerBackgroundRadius
    real(single)     :: outerBackgroundRadius
    real(single)     :: tdsCorr     
    real(double) :: pa   
    type(OmProgressT)     :: reporter    ! ... of progress to GUI    
    type(ColumnT) :: columnh
    integer i
      ! --- don't bother if none available
    if (trim(setName) == "" ) return




!write(*, *) "omsource_old line 1449"



    ! Read in the OM OSW Source List file.
    set = dataSet(setName, read)
    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  


!write(*, *) "omsource_old line 1458, file=", setName



    if (.not. hasBlock(set, "SRCLIST")) then

!        call destruct(reporter)

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else
        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")
!write(*, *) "omsource_old line 1473"

    end if

     
!write(*, *) "omsource_old line 1478"

    ! Get the limiting magnitude of the image
    myDetectionRate      = GetReal32Attribute(set, "CRLIMIT")
    myDetectionMagnitude = GetReal32Attribute(set, "MAGLIMIT")



!write(*, *) "omsource_old line 1473"

    
    innerBackgroundRadius =  GetReal32Attribute(set, "INNBCKAN")
    outerBackgroundRadius =  GetReal32Attribute(set, "OUTBCKAN")

!write(*, *) "omsource line 1351 innerradius=", innerBackgroundRadius, " outer=", outerBackgroundRadius  



    if (hasAttribute(inTableH, "TDS_CORR")) then
        tdsCorr =  real32Attribute(inTableH, "TDS_CORR")
    end if


    if (innerBackgroundRadius .eq. 0.0) then

        innerBackgroundRadius = myBgrRadiusIn
        outerBackgroundRadius =  myBgrRadiusOut

    end if
 
    call construct(reporter, label="Loading sourcelist data", steps = 23)


!write(*, *) "omsource_old line 1497"


    ! Get the number of rows in the source list binary table extension.
    nSources = numberOfRows(inTableH)

    if (nSources == 0) then
        call destruct(reporter)

        call release(inTableH)
        call release(set)
        return

    end if

    call delete(sources)
    call new(sources, low = 1, high = nSources)
 
    ! Get the data for each source from the source list file
    call setProgress(reporter, 1)
  
    call LoadReal32Column(inTableH, "XPOS",  sources%xPos)
    call setProgress(reporter, 2)
  
    call LoadReal32Column(inTableH, "YPOS",  sources%yPos)
    call setProgress(reporter, 3)
  
    call LoadReal32Column(inTableH, "POSERR",  sources%poserr)
    call setProgress(reporter, 3)
  

!write(*, *) "omsource_old line 1528"



    call LoadReal32Column(inTableH, "FWHM_MAJ",  sources%semiA)
    call setProgress(reporter, 4)
  
    call LoadReal32Column(inTableH, "FWHM_MIN",  sources%semiB)
    call setProgress(reporter, 5)
  
    call LoadReal32Column(inTableH, "PA", sources%theta)
    call setProgress(reporter, 6)
  
    call LoadReal32Column(inTableH, "RATE", sources%rawCountRate)
    call setProgress(reporter, 7)
  
    call LoadReal32Column(inTableH, "RATE_ERR", sources%rawCountRateErr)
    call setProgress(reporter,8)

    call LoadReal32Column(inTableH, "CORR_RATE", sources%corrCountRate)
    call setProgress(reporter, 9)
  
    call LoadReal32Column(inTableH, "CORR_RATE_ERR", sources%corrCountRateErr)
    call setProgress(reporter, 10)

    call LoadReal32Column(inTableH, "BK_COILOSS", sources%semiaErr)

    call LoadReal32Column(inTableH, "SPB_COILOSS", sources%semibErr)





    call setProgress(reporter,11)
    call LoadReal32Column(inTableH, "PSF1_CORR", sources%psf1Correction)
    call setProgress(reporter, 12)
  
    call LoadReal32Column(inTableH, "PSF2_CORR", sources%psf2Correction)
    call setProgress(reporter,13)




    call LoadReal32Column(inTableH, "BACKGROUND_RATE",  sources%rawBackgroundCountRate)
    call setProgress(reporter, 14)
  
    call LoadReal32Column(inTableH, "CORR_BACKGROUND_RATE",  sources%corrbackgroundCountRate)
    call setProgress(reporter,15)

    call LoadReal32Column(inTableH, "MAG", sources%magnitude)
    call setProgress(reporter, 16)
  
    call LoadReal32Column(inTableH, "MAGERR", sources%magnitudeErr)
    call setProgress(reporter, 17)
  
    call LoadReal32Column(inTableH, "SIGNIFICANCE", sources%significance)
    call setProgress(reporter, 18)
  
    call LoadReal32Column(inTableH, "APERTURE", sources%apertureRadius)


    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)


!write(*, *) "omsource_old line 1592"


    !**************************************************************
    ! Find out the data type
    !************************************************************
    columnH = Column(inTableH, "QFLAG")




    if( dataType(columnh) .eq. integer16) then 
        call LoadInt16Column(inTableH, "QFLAG", sources%qualityFlag)
    else
        call LoadInt8Column(inTableH, "QFLAG", sources%confusionFlag)
        sources%qualityFlag = int(sources%confusionFlag, int16)


    end if


    call LoadInt8Column(inTableH, "CFLAG", sources%confusionFlag)
 


!write(*, *) "omsource_old line 1617"



    ! Need binned aperture radius
    where(sources%extensionflag .eq. 1) 
        sources%apertureRadius = 0.0
    elsewhere
        sources%apertureRadius = sources%apertureRadius / myImageWindow%bx
    endwhere
    call setProgress(reporter, 19)

    ! Get the RAS and decs if they exist
    call LoadReal64Column(inTableH, "RA", sources%rightAscension)
    call setProgress(reporter, 20)
  
    call LoadReal64Column(inTableH, "DEC", sources%declination)
    call setProgress(reporter, 21)
  

!write(*, *) "omsource_old line 1637"



    call setProgress(reporter,22)

  

    sources%innerBckRadius = innerBackgroundRadius

    sources%outerBckRadius = outerBackgroundRadius
    sources%exposure = myExposure 


!write(*, *) "omsource-old line 1718 myexposure=", myExposure

    sources%tdsCorr = tdsCorr
    ! Convert width to arcsecs
  
    sources%semiA = sources%semiA / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)
    sources%semiB = sources%semiB / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)



!write(*, *) "omsource_old line 1661"

   
    if (hasColumn(inTableH, "RA")) then
        pa = real64Attribute(set, "PA")
        if( myExposure > 2.0) then
            sources%theta = sources%theta - 90.0 - PA
            where(sources%theta .lt. 0.0) sources%theta = sources%theta + 180.0 
            where(sources%theta .lt. 0.0) sources%theta = sources%theta + 180.0 
        end if


    end if


!write(*, *) "omsource_old line 1676"


    sources%alpha = sources%theta
 
    do i = 1, size(sources)

        sources(i)%number = i

!write(*, *) i, ") x=", sources(i)%xpos, " y=", sources(i)%ypos, " a=", sources(i)%semia, &
!" b=", sources(i)%semib, " theta=", sources(i)%theta
   end do
         
    call setProgress(reporter,23)

    call destruct(reporter)





    call release(inTableH)
    call release(set)


!write(*, *) "omsource_old line 1701"


end subroutine loadSources

!**********************************************************
! Info: saves the sources into dataset(fileName)
! Used globals:
!   mySourcesAreLoaded
!   myDetectionRate
!   myDetectionMagnitude
!   myFrameTime
!   myDeadFraction
!   myNullValue
! TODO: save unbinnend radius? save baground radii?
!*********************************************************
subroutine omsource_save(sources, fileName, imageSet, iw)

    type(SourceParameters), pointer    :: sources(:)
    character(*),            intent(in) :: fileName
    type(DataSetT),          intent(in) :: imageSet
    type(ImageWindow),       intent(in) :: iw


    integer                :: nSources

    type(OmProgressT)      :: reporter    ! ... of progress to GUI
    character(*) , parameter :: routine="omsource_old.omsource_save"

    if (.not.associated(sources)) return ! don't bother
    nSources = size(sources)

    sources%apertureRadius = sources%apertureRadius *  real(myImageWindow%bx) 

    ! notify the GUI that we are busy...
    call construct(reporter, label="Saving source list- please wait", steps = 2)
    call setProgress(reporter, 1)



    ! Create the source-list table
    call CreateSourceListFile(fileName,         &
                              imageSet,         &
                              nSources,         &
                              sources,          &
                              iw%plateScale,    &
                              myApertureRadius / real(myImageWindow%bx) , &
                              myBgrMode,        &
                              3.0,              &
                              myExposure,       & 
                              iw%x0,            &
                              iw%y0,            & 
                              iw%width,         &
                              iw%height) 
    call AddMagnitudesToTable(fileName, sources)     
    call setProgress(reporter, 2)

    call destruct(reporter)

end subroutine omsource_save

!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!****************************************************************
subroutine computeParameters(sources)
    
    type(SourceParameters),pointer    :: sources(:)

    integer(int32)                     :: nSources,nx, ny
   
    
   
    nx = size(myimage, 1)
    ny = size(myimage, 2)
    nSources = size(sources)

    
    call omsource_ComputeSourcePositions(sources)


    call omsource_photometry(myimage, nSources, sources) ! see below
    

    ! Compute the RA and Declination of each source
    call setCelestial(myCelestial, sources)
end subroutine computeParameters



!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!       * 'trimage' argument is a residual, and should be removed
!****************************************************************
subroutine omsource_ComputeSourcePositions(sources)
    
    type(SourceParameters),pointer    :: sources(:)

    integer(int16)                     :: i,nSources,nx, ny
    integer(int32)                     :: i0, j0
    integer(kind=int16) :: sourceNumber = 1
   
    nx = size(myImage, 1)
    ny = size(myImage, 2)

    nSources = size(sources)

   
!write(*, *) "Compute source positions"

    ! Compute the centre coordinates of the new sources

    do i = 1, nSources

!      write(*, *) i, " before x=", sources(i)%xPos, " y=", sources(i)%yPos 
       
        call computeCentreOfSource(nx,ny, &
              sources(i)%xPos, sources(i)%yPos, &
              i0, j0)

!  write(*, *) i, " after computecentreofsource x=", sources(i)%xPos, " y=", sources(i)%yPos 

        sources(i)%iMin = max(i0 - 3, 0)
        sources(i)%iMax = min(i0 + 3, nx - 1)
        sources(i)%jMin = max(j0 - 3, 0)
        sources(i)%jMax = min(j0 + 3, ny - 1)

        call ComputePointSourceParameters(sourceNumber, nx - 1, ny - 1, myImage, &
                i0, j0, sources(i))

!      write(*, *) i, " after x=", sources(i)%xPos, " y=", sources(i)%yPos 

!write(*, *) "imin=", sources(i)%iMin, " max=",  sources(i)%iMax
!write(*, *) "jmin=", sources(i)%jMin, " max=",  sources(i)%jMax
 
    end do

    
end subroutine omsource_ComputeSourcePositions



!************************************************************
! Subroutine photometry
!***********************************************************
subroutine omsource_photometry(image, nSources,  sources)
    real(single),           pointer    :: image(:, :)
    integer(int32),         intent(in) :: nSources
    type(SourceParameters),pointer    :: sources(:)
 
    integer(int32)         :: i
    integer(int16)         :: nx, ny
    integer(int32)         :: bin(2)
    type(SourceParameters), pointer :: s 

    real(single) :: corrBackgroundCountRateErr
    type(SourceParameters),pointer    :: tempsources(:) => null()
  

    nx = size(image, 1)
    ny = size(image, 2)


!write(*, *) "omsource-old line 1920calling aperture photometry, nsources=", nSources
!write(*, *) "exposure=", myObsDetails%exposure




    ! initialise defaults:
    do i = 1, nSources
        s=>sources(i)
     
        s%innerBckRadius = myBgrRadiusIn
        s%outerBckRadius = myBgrRadiusOut
        s%exposure       = myObsDetails%exposure
        
        s%extensionFlag  = 0 ! Assume all sources are "point-like"


!write(*, *) i, ") x=", sources(i)%xpos, " y=", sources(i)%ypos, &
!        " aperture radius=", sources(i)%apertureradius, &
!        " rawbackgroundcountrate=", sources(i)%rawbackgroundCountRate
!write(*, *) "inner background radius=", sources(i)%innerbckradius, &
!            "outer background radius=", sources(i)%outerbckradius

    end do

!write(*, *) "My aperture radius-", myApertureRadius / real(binx)

    ! Do aperture photometry on each sources
    bin = (/myImageWindow%bx, myImageWindow%by/)
 
    allocate( tempSources(1 : nsources))
    tempSources = sources(1 : nsources)


!write(*, *) "omsource-old line 1951 calling aperture photometry"
!write(*, *) "coinclossradius=", coinclossradius


    call AperturePhotometry(int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
          myImage, mylevelImage,                                                        &
          myQuality, myapertureRadius / real(binX),  int(nSources, int16), sources)                  

    sources(1 : nsources)%semiberr = sources%rightascension
    sources(1 : nsources)%semiaerr = sources%declination
    sources(1 : nsources)%rightAscension = tempSources%rightAscension
    sources(1 : nsources)%declination = tempSources%declination

    
    deallocate(tempSources)
    nullify(tempSources)

    ! Convert counts to count rates, and set the magnitude
    do i = 1, nSources
        s=>sources(i)

!write(*, *) "omsource_old line 1981 BEFORE calling compute magnitude - s%corrcountrate=",s%corrCountRate 
!write(*, *) "s%corrBackgroundCountRate=",  s%corrBackgroundCountRate
!write(*, *) " frametime=", myFrameTime, " deadfraction=", myDeadFraction
!write(*, *) "apertureradius=", myApertureRadius
!write(*, *) " tdscorr=",s%tdsCorr, " exposure=", s%exposure
!write(*, *) "x=", s%xpos, " y=", s%ypos

        call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr,               &
                          s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                          s%xPos, s%yPos,                                        &
                          myScaleCounts, s%extensionFlag,                        &
                          myCalRadius, myApertureRadius,                         &
                          myFrameTime, myDeadFraction,                           &
                          s%magnitude, s%magnitudeErr, s%psf2Correction,         &
                          myobsDetails%filter, myObsDetails%mjd, s%tdsCorr)


!write(*, *) "mag=", s%magnitude, " error=", s%magnitudeErr

    !************************************************************************
    ! NB after calling aperture photometry the aperture radius will have been
    ! multiplied by the binning factor binx
    !************************************************************************


    s%apertureRadius = s%apertureRadius / real(binX)


!write(*, *) "omsource-old line 2013 after compute magnitude deadfraction=", myDeadFraction
!write(*, *) "s%corrcountrate=",s%corrCountRate 
!write(*, *) "s%corrBackgroundCountRate=",  s%corrBackgroundCountRate
!write(*, *) " frametime=", myFrameTime, " deadfraction=", myDeadFraction
!write(*, *) "apertureradius=", myApertureRadius
!write(*, *) " tdscorr=",s%tdsCorr


    end do

end subroutine omsource_photometry
!************************************************************

   ! Info: find a more accurate centre pixel for (xPos,yPos)
!***********************************************************
subroutine computeCentreOfSource(nx, ny, xPos, yPos, i0, j0)
    integer(int16), intent(in) :: nx, ny

     
    real(single), intent(in) :: xPos, yPos

    integer(int32), intent(out) :: i0, j0
    integer(int32) :: il, iu, jl, ju

    integer(int32) :: i, j
    real(single) :: maximum

      ! current pixel centre

    i0 = int(xPos - 0.5)
    j0 = int(yPos - 0.5)

    ! now find a more accurate centre for the image
    il = i0 - 5
    il = max (il, 0)
    iu = i0 + 5
    iu = min (iu, nx - 1)

    jl = j0 - 5
    jl = max (jl, 0)
    ju = j0 + 5
    ju = min (ju, ny - 1)
    maximum = -1e10

    do i = il, iu
        do j = jl, ju
            if (myImage(i, j) .gt. maximum) then
                maximum = myImage(i, j)
                i0 = i
                j0 = j
            end if
        end do
    end do


end subroutine computeCentreOfSource
!***************************************
!
!***************************************
subroutine construct_od(o, set)
    type(ObsDetails),  intent(out)    :: o
    type(DataSetT),    intent(in)     :: set
      
    call GetStringAttribute(set, "TELESCOP", o%telescope)
    call GetStringAttribute(set, "OBSERVER", o%observer)
    call GetStringAttribute(set, "OBJECT",   o%object)
    call GetStringAttribute(set, "FILTER",   o%filter)
    call GetStringAttribute(set, "INSTRUME", o%instrument)
    call GetStringAttribute(set, "OBS_ID",   o%obsID)
    call GetStringAttribute(set, "EXP_ID",   o%expID)
    call GetStringAttribute(set, "DATE-OBS", o%startObs)
    call GetStringAttribute(set, "DATE-END", o%endObs)
    o%exposure = GetReal32Attribute(set, "EXPOSURE")

!write(*, *) "omsource-old line 2093 construct- exposure=", o%exposure


    !*************************************************************************
    ! Need the modified Julian date for time-dependent sensitivity correction
    !************************************************************************
     o%mjd = ModifiedJulianDate(trim(o%startObs))



end subroutine construct_od
!***************************************
!
!***************************************
subroutine destruct_od(o)
    type(ObsDetails), intent(inout)    :: o
    o%telescope=''
    o%exposure=0.
end subroutine destruct_od
!***************************************
!
!***************************************
subroutine construct_iw(w, set)
    type(ImageWindow), intent(out)    :: w
    type(DataSetT),    intent(in)     :: set
      
    type(TableT)                      :: t
    type(ColumnT)                     :: c
    integer(int32), pointer            :: ptr(:) ! release on table release
    type(DataSetT) :: wdxDataSet      

   
    if( hasBlock(set, "MODES") ) then
        t = table(set, "MODES")
        c = column(t, "WINDOWX0", read)
        ptr => int32Data(c)
        w%x0  = ptr(0)
        c = column(t, "WINDOWY0", read)
        ptr => int32Data(c)
        w%y0  = ptr(0)
        c = column(t, "WINDOWDX", read)
        ptr => int32Data(c)
        w%width  = ptr(0)
        c = column(t, "WINDOWDY", read)
        ptr => int32Data(c)
        w%height  = ptr(0)
        call release(t)
    else
        w%x0 = int32Attribute(set, "WINDOWX0")
        w%y0 = int32Attribute(set, "WINDOWY0")
        w%width = int32Attribute(set, "WINDOWDX")
        w%height = int32Attribute(set, "WINDOWDY")


    end if

    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  
   

   

    ! compute binning
    w%binBpe = booleanAttribute(set, "BINBPE")
    w%bx     = 2 ** int32Attribute(set, "BINAX1")
    w%by     = 2 ** int32Attribute(set, "BINAX2")
    if (w%binBpe) then
        w%bx = w%bx * 2
        w%by = w%by * 2
    end if


    binX = w%bx
    binY = w%by

    w%filter  = stringAttribute(set, "FILTER")
    w%datamode= stringAttribute(set,"DATAMODE")


    !*************************************************
    ! CCD frame-time and dead-fraction
    !*************************************************
    
    if (.not. hasAttribute(set,  "FRAMTIME")) then
        wdxDataSet = dataSet(stringParameter("wdxset"), read)


        call getFrameTime(wdxDataSet, w%frameTime, w%deadFraction, w%binBpe)
        call release(wdxDataSet)
        call setAttribute(set, "FRAMTIME", w%FrameTime, "millisecs", "CCD frame time")
        call setAttribute(set, "DEADFRAC", w%DeadFraction, "", "CCD dead fraction")
        w%frameTime = w%frameTime / 1000.0
       
    else     
        w%frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
        w%deadFraction = real64Attribute(set, "DEADFRAC")

    end if 

end subroutine construct_iw
!***************************************
!
!***************************************
subroutine destruct_iw(w)
    type(ImageWindow), intent(inout)    :: w
    w%filter=''
    w%datamode=''
end subroutine destruct_iw



subroutine omsource_computeparameters(sources)
    type(SourceParameters),  pointer    :: sources(:)

    call ComputeParameters(sources)
end subroutine omsource_computeparameters


!*************************************************
!
! Returns the background mode
!*************************************************

real(single) Function omsource_GetBackgroundMode()
 omsource_GetBackgroundMode = myBgrMode
    
end function omsource_GetBackgroundMode


!****************************************************************************
! SetTheSourceFlags
!***************************************************************************
subroutine SetTheSourceFlags(sources)
    type(SourceParameters),pointer    :: sources(:)
 

    type(SourceParameters),pointer    :: tempsources(:) => null()
    integer :: nsources 


    if (.not. associated(sources)) return
    nSources = size(sources)

    


    allocate( tempSources(1 : nsources))
    tempSources = sources(1 : nsources)





    call setsourceflags(int(nsources, int16), sources, myimagewindow%x0, myimagewindow%y0, & 
         mylevelimage) 

    sources%xpos = tempSources%xpos
    sources%ypos = tempSources%ypos


    deallocate(tempSources)
    nullify(tempSources)



end subroutine SetTheSourceFlags

!*********************************************
!
! subroutine GetExtendedSourceCounts1
! For extended sources with elliptical apertures
! Computes the count-rates/etc for an extended source
! 
!*********************************************
subroutine OmsourceExtendedSourceCounts1( image, source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer :: image(:, :)
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta
      
    real(double) :: f, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea
    real(single) :: error    
    logical :: useImage
    integer :: imin, imax, jmin, jmax  
     
    ccdimage => image
 


!write(*, *) "OmsourceExtendedSourceCounts1 line 2208 - rawbackgroundcountrate=",  source%rawBackgroundCountRate


    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     

    call initialiseArrays

!image = 10.0
 
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    ! the elliptical background annulus

!write(*, *) "getextendedsourcecounts1 semia=", semia, " semib=", semib, " angle=", angle
!write(*, *) "inner bckradius=",   source%innerBckRadius, " outer=", source%outerBckRadius, " axialratio=",source%axialRatio  
    if (source%rawBackgroundCountRate > 0.0) then ! Using a fixed background mode to get background counts
        backgroundCounts = source%rawBackgroundCountRate * PI * semiA * semiB
        useImage = .false.


    else ! Using the mode from the image
        
        backgroundCounts = GetBackGroundMode2( &
                     source%xPos,              &   ! The centre x-pixel coordinate of the aperture
                     source%yPos,              &   ! The centre y-pixel corrdinate of the aperture 
                     source%innerBckRadius,    &   ! The aperture inner radius
                     source%outerBckRadius,    &   ! The aperture outer radius
                     source%axialRatio,        &   ! The ellipse axial ratio
                     theta,                    &   ! The ellipse position angle
                     error)      ! The error
        useImage = .true.

    end if    

!write(*, *) "Line 2259 backgroundCounts=", backgroundCounts

    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
    totalArea = 0.0_double



 

    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 



    meanExposure = 0.0_double
    numExposurePixels = 0
    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
            if( isNullValue( image(i, j)) ) then
                cycle

            end if       
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
 
            if(associated(exposureImage)) then
                meanExposure = meanExposure + dble(myExposureImage(i, j))
                numExposurePixels = numExposurePixels + 1
            end if 


           totalArea = totalArea + pixArea
            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 !           image(i, j) = 100.0
            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
         
            else
                f =  PixelValue2(i, j)            
          
            end if

            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), myframeTime)
                                 
        end do
    end do

!call SaveRealImage(image, int(nx, int16) + 1_int16, int(ny, int16) + 1_int16, "omsourceimage.fits", "test")


    if (useImage) backgroundCounts = backgroundCounts * totalArea

!write(*, *) "semia=", semia, " semib=", semib, " axialratio=", source%axialratio
!write(*, *) "total area=", totalarea, " theoretical=", (PI * semia * semib)
!write(*, *) "total counts=", rawCounts, " backgroundcounts=", backgroundcounts
    backgroundRate = backgroundCounts / source%exposure ! Background count rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      





    !**********************************************************
    ! The coincidence-loss correction for the background
    !*********************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), myframeTime)
    source%corrBackGroundCountRate = backgroundRate * correction ! Corrected background rate


    correctedBackgroundCounts = backgroundCounts * correction

   
   

    ! Background-subtracted counts
    rawCounts = rawCounts - backgroundCounts
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 


    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               


    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts) / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 




          
    source%psf1Correction = 0.0



    ! The source coincidence-loss correction
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  source%corrCountRate / source%rawCountRate
    end if
    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  source%corrBackGroundCountRate / source%rawBackGroundCountRate
    end if



    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
    nullify(ccdimage)




    deAllocate(backgroundPixel)
    nullify(backgroundPixel)





end subroutine OmsourceExtendedSourceCounts1      



!*********************************************
! subroutine GetExtendedSourceCounts2
! For extended sources with elliptical apertures
! Computes the count-rates/etc for an extended source
! 
!*********************************************
subroutine OmsourceExtendedSourceCounts2( image, source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer :: image(:, :)
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, fb, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
    real(single), pointer, dimension(:) :: x=>null(), y=>null(), counts=>null()
    integer :: ncount
          

    integer, parameter :: maxSize = 10000
    real(single) :: innerbckSemia, outerbcksemia
    real(single) :: innerbckSemib, outerbcksemib
    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


!write(*, *) "OmsourceExtendedSourceCounts2 - interpolating using annulus pixels"

    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
     



      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     





    !***************************************************************************
    ! Set up a background interpolant using the pixels in the background annulus
    !************************************************************************** 

    x => allocateSingle( 1, maxSize, "x")
    y => allocateSingle( 1, maxSize, "y")
    counts => allocateSingle( 1, maxSize, "counts")


    innerbcksemia = source%innerBckRadius * source%axialRatio
    outerbckSemia = source%outerBckRadius * source%axialRatio
    innerbcksemib =  source%innerBckRadius
    outerbcksemib =  source%outerBckRadius

    ncount = 0
    imin =  max(0, int(floor(source%xPos) - ceiling(outerbcksemiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + outerbcksemiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(outerbcksemiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + outerbcksemiA + 0.1) + 1) 
 


iloop :    do i = imin, imax 
        x0 = real(i + 1) 
     
        do j = jmin, jmax
            if(isNullValue( image(i, j)) ) then
                cycle

            end if       
                            
        
                                  
            y0 = real(j + 1) 
      
            !********************************************************************
            ! Point has to be within the inner and outer ellipse
            !********************************************************************
            if (.not. PointInEllipse(x0, y0, source%xpos, source%ypos, innerbcksemiA, innerbcksemiB, thetar) .and. &
                      PointInEllipse(x0, y0, source%xpos, source%ypos, outerbcksemiA, outerbcksemiB, thetar)) then

           
          
     
              
                ncount = ncount + 1
                x(ncount) = real(i) + 0.5
                y(ncount) = real(j) + 0.5
                    
                counts(ncount) = image(i, j)  
                if (ncount == maxSize) then

                  exit iLoop
                end if  
            end if
                    
        end do    
    end do iLoop

 

   
    backgroundCounts = 0.0_double
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
 


    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 


    meanExposure = 0.0_double
    numExposurePixels = numExposurePixels + 1

    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
       
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea



            if(associated(exposureImage)) then
                meanExposure = meanExposure + dble(myExposureImage(i, j))
                numExposurePixels = numExposurePixels + 1
            end if 




            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            !***************************************************************************
            ! Interpolate to get the background counts
            !**************************************************************************
            fb = neighbourInterpolate( x0, y0, ncount, x, y, counts ) 
          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
                fb = fb * pixarea 
            else
                f =  PixelValue2(i, j)            
              
            end if
            backgroundCounts = backgroundCounts + fb
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), myframeTime)
          
!image(i, j) = fb                       
        end do
    end do


!call SaveRealImage(image, int(nx, int16) + 1_int16, int(ny, int16) + 1_int16, "omsourceimage.fits", "test")


!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
!write(*, *) "coinclossarea=", coinclossArea, " backgroundcounts=", backgroundCounts, " totalarea=", totalarea


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate



    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), myframeTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
!write(*, *) "exposure=", source%exposure
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 
!write(*, *) "rawcounts=", rawCounts


    ! The source coincidence-loss correction
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               
    
    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

!write(*, *) "rawcounts=", rawCounts, " backgroundcounts=", backgroundCounts 
!write(*, *) "rawcountrate=", source%rawCountRate, " background=", source%rawbackgroundCountRate
     

    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
    deallocate(x, y, counts)
    nullify(x, y, counts)




end subroutine OmsourceExtendedSourceCounts2      





!*********************************************
! subroutine GetExtendedSourceCounts3
! For extended sources with elliptical apertures
! and a fixed background value
! Computes the count-rates/etc for an extended source
! 
!*********************************************
subroutine OmsourceExtendedSourceCounts3( image, source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer :: image(:, :)
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
           

    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


!write(*, *) "OmsourceExtendedSourceCounts3 - interpolating using fixed background value"

    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
     



      
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     








   
    backgroundCounts = 0.0_double
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
 


    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 

    meanExposure = 0.0_double
    numExposurePixels = 0

    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
            if(isNullValue( image(i, j) )) then
                cycle

            end if       
                            
        
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea

            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            if(associated(exposureImage)) then
                meanExposure = meanExposure + dble(myExposureImage(i, j))
                numExposurePixels = numExposurePixels + 1
            end if 


          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
          
            else
                f =  PixelValue2(i, j)            
              
            end if
           
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), myframeTime)
          
!image(i, j) = fb                       
        end do
    end do


    backgroundCounts = source%rawBackgroundCountRate * totalArea



!call SaveRealImage(image, int(nx, int16) + 1_int16, int(ny, int16) + 1_int16, "omsourceimage.fits", "test")


!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
  


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate



    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), myframeTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 



    ! The source coincidence-loss correction
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

      
    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
  




end subroutine OmsourceExtendedSourceCounts3      















!*******************************************************************************
!
!*******************************************************************************
subroutine omsourceold_recomputephotometry(sources, numberToFit, no, type, radius, &
      innerA, outerA, axialRatio, theta, background)
    type(SourceParameters), pointer :: sources(:)

    integer, intent(in) :: numberToFit
    integer, intent(in) :: no(1 : numberToFit)

    
    integer, intent(in) :: type
    real(double), intent(in) :: radius, innerA, outerA, axialRatio, theta, background

    real(single) :: corrBackgroundCountRateErr = 0.0
    type(SourceParameters), pointer  :: temp(:) => null()
    type(SourceParameters) :: s
    integer(int8) :: i


    ! Do aperture photometry on  source

    allocate(temp( 1 : numberToFit))



!write(*, *) "omsource-old Line 2972 recompute photometry numbertofit=", numberToFit, " type=", type

    do i = 1, numberToFit
!        write(*, *) i, ") no=", no(i), " number=", sources(no(i))%number, &
!          " x=", sources(no(i))%xpos, " y=", sources(no(i))%ypos
        temp(i) = sources(no(i))

        temp(i)%apertureradius = radius
        temp(i)%innerbckradius = innerA
        temp(i)%outerbckradius = outerA
        temp(i)%exposure = myExposure
        temp(i)%axialRatio = axialRatio
        temp(i)%theta = theta
        temp(i)%rawCountRate = -1.0
        temp(i)%rawBackgroundCountRate = -1.0

    

    end do     


    if (.not. associated(myImage)) then
        call fatal("myimage not associated")

    end if


    if (.not. associated(mylevelImage)) then
        call fatal("mylevelimage not associated")

    end if


    if (.not. associated(myquality)) then
        call fatal("myquality not associated")

    end if


    !************************************************************************************
    !
    ! Point-source photometry if type equals 1
    !
    !************************************************************************************ 
    select case (type)

       case (0)  ! DEFAULT APERTURE PHOTOMTERY
            temp(1)%extensionFlag = 0


            !*************************************************
            ! Aperture photometry
            !*************************************************
            write(errstr,'(a)') "Computing photometry using circular aperture"
            call message(AppMsg, Verbose)


            if( fastBackground) then

                !*************************************************************
                ! For FAST data might be using a bigger image to get the
                ! background
                !*************************************************************
                call getFastBackground( temp(1))
    
                call AperturePhotometry(int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
                    myImage, mylevelImage,                                                          &
                    myQuality, myapertureRadius / real(binX),                                       &  
                    1_int16, temp)
        
            else    

                call AperturePhotometry(int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
                    myImage, myBackgroundImage, mylevelImage,                                       &
                    myQuality, myapertureRadius / real(binX),                                       &  
                    1_int16, temp)
            end if 


        case (1)  ! DEFAULT APERTURE PHOTOMTERY- FIXED BACKGROUND
            temp(1)%extensionFlag = 0
            temp(1)%rawBackgroundCountRate = background
   

            ! Aperture photometry
            write(errstr,'(a, f6.1)') &
                 "Computing photometry using circular aperture- background counts per pixel=", background
            call message(AppMsg, Verbose)

  
            call AperturePhotometry(int(size(myImage, 1), int16), int(size(myImage, 2), int16), &
                    myImage, mylevelImage,                                                      &
                    myQuality, myapertureRadius / real(binX),                                   &  
                    1_int16, temp)
               
        case(2)   ! Least-squares fit to one or more sources, including background

            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from A PSF least-squares fit which includes " &
                // "a constant term for the background"
            call message(AppMsg, Verbose)


            call PSFPhotometry3(int(numberToFit, int8), myimage, temp)
 
        case(3)   ! Least-squares fit to one or more sources, fixed background

            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from A PSF least-squares fit which includes " &
                // "a constant term for the background"
            call message(AppMsg, Verbose)
!write(*, *) "omsource_old.f90 line 3089 calling psfphotometry3"

            call PSFPhotometry33(int(numberToFit, int8), myimage, temp, real(background))
     
        case(4)   ! Least-squares fit to one or more sources, including background and positions

            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from A PSF least-squares fit which includes " &
                // "a constant term for the background and a fit to the positions"
            call message(AppMsg, Verbose)


            call PSFPhotometry4(myImage, temp, int(numberToFit, int32))

 
        case(5)   ! Least-squares fit to one or more sources, fixed background

            temp%extensionFlag = 0

            ! Write out the source list to the oswDataSet
            write(errstr,'(a)') "Computing photometry from A PSF least-squares fit which includes " &
                // "a constant term for the background, but a fit to the positions"
            call message(AppMsg, Verbose)

            call PSFPhotometry44(myImage, temp, int(numberToFit, int32), real(background))

        case default
            call fatal("UNKNOWN PHOTOMETRY TYPE!")
    end select  


    !************************************************************************************
    ! Compute the magnitude
    !
    !************************************************************************************
  
    s = temp(1)  

!write(*, *) "omsource_old line 3154 BEFORE calling compute magnitude - s%corrcountrate=",s%corrCountRate 
!write(*, *) "s%corrBackgroundCountRate=",  s%corrBackgroundCountRate
!write(*, *) " frametime=", myFrameTime, " deadfraction=", myDeadFraction
!write(*, *) "line 446 apertureradius=", myApertureRadius, " source aperture radius=", s%apertureRadius
!write(*, *) " tdscorr=",s%tdsCorr
  


    call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr,                   &
                          s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                          s%xPos, s%yPos,                                        &
                          myScaleCounts, s%extensionFlag,                        &
                          myCalRadius, myApertureRadius,                         &
                          myframeTime, mydeadFraction,                           &
                          s%magnitude, s%magnitudeErr, s%psf2Correction,         &
                          trim(myObsDetails%filter) , myObsDetails%mjd , s%tdsCorr)

!write(*, *) "Computed magnitude=", s%magnitude

    !************************************************************************
    ! Set the source in the sourcelist to the new one
    !***********************************************************************  
    sources(no(1)) = s



   
    deallocate(temp)



!write(*, *) "omsource-old line 3295 aperture photometry binx=", binx


end subroutine omsourceold_recomputephotometry






!**************************************************************************
! 
! For extended sources with elliptical apertures
! semia = semi-major axis of aperture
! semib = semi-minor axis of aperture
! angle is the position angle of the ellipse
! source%innerBckRadius = semi-major axis of inner background annulus
! source%outerBckRadius = semi-major axis of outer background annulus
! 
!**************************************************************************
subroutine ExtendedSourcePhotometry1( image, backgroundImage, &
                source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer, dimension(:, :) :: image, backgroundimage
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, fb, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
    real(single), pointer, dimension(:) :: x=>null(), y=>null(), counts=>null()
    integer :: ncount
          

    integer, parameter :: maxSize = 10000
    real(single) :: innerbckSemia, outerbcksemia
    real(single) :: innerbckSemib, outerbcksemib
    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
           
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     


    !***************************************************************************
    ! Set up a background interpolant using the pixels in the background annulus
    !************************************************************************** 

    x => allocateSingle( 1, maxSize, "x")
    y => allocateSingle( 1, maxSize, "y")
    counts => allocateSingle( 1, maxSize, "counts")


    innerbcksemia = source%innerBckRadius
    outerbckSemia = source%outerBckRadius
    innerbcksemib = source%innerBckRadius / source%axialRatio
    outerbcksemib = source%outerBckRadius / source%axialRatio

    ncount = 0
    imin =  max(0, int(floor(source%xPos) - ceiling(outerbcksemiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + outerbcksemiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(outerbcksemiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + outerbcksemiA + 0.1) + 1) 
 


iloop :    do i = imin, imax 
        x0 = real(i + 1)  
    
        do j = jmin, jmax
                                  
            y0 = real(j + 1) 
      
            !********************************************************************
            ! Point has to be within the inner and outer ellipse
            !********************************************************************
            if (.not. PointInEllipse(x0, y0, source%xpos, source%ypos, innerbcksemiA, innerbcksemiB, thetar) .and. &
                      PointInEllipse(x0, y0, source%xpos, source%ypos, outerbcksemiA, outerbcksemiB, thetar)) then


           
                                 
                ncount = ncount + 1
                x(ncount) = real(i) + 0.5
                y(ncount) = real(j) + 0.5
                    
                counts(ncount) = backgroundimage(i, j)  
     
 !              image(i, j) = -1.0          
    
                if (ncount == maxSize) then
                      
                  exit iLoop
                end if  
            end if
                    
        end do    
    end do iLoop

 

   
    backgroundCounts = 0.0_double
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
 


    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 

    meanExposure = 0.0_double
    numExposurePixels = 0

    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
       
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)

            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea


            if(associated(exposureImage)) then
                meanExposure = meanExposure + dble(myExposureImage(i, j))
                numExposurePixels = numExposurePixels + 1
            end if 






            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            !***************************************************************************
            ! Interpolate to get the background counts
            !**************************************************************************
            fb = neighbourInterpolate( x0, y0, ncount, x, y, counts, 4.0 * semiB ) 
          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
                fb = fb * pixarea 
            else
                f =  PixelValue2(i, j)            
              
            end if
            backgroundCounts = backgroundCounts + fb
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), myframeTime)
          
!image(i, j) = fb                       
        end do
    end do


!call SaveImage(image, "omsourceimage.fits", "test")








!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
  
!write(*, *) "exposure=", exposure, " coinclossarea=", coinclossarea 


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate

!write(*, *) "background counts=", backgroundcounts, " backgroundrate=", backgroundrate


    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), myframeTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 


    !****************************************************
    ! The source coincidence-loss correction
    !****************************************************
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

      
    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
    deallocate(x, y, counts)
    nullify(x, y, counts)

end subroutine extendedSourcePhotometry1      



!**************************************************************************
! 
! For extended sources with elliptical apertures
! semia = semi-major axis of aperture
! semib = semi-minor axis of aperture
! angle is the position angle of the ellipse
! source%innerBckRadius = semi-major axis of inner background annulus
! source%outerBckRadius = semi-major axis of outer background annulus
! 
!**************************************************************************
subroutine ExtendedSourcePhotometry2( image, &
                source, coincLossRadius, semiA, semiB, angle)
    real(single), pointer, dimension(:, :) :: image
    ! Input (unmodfied) variables
    real(single), intent(in) :: coincLossRadius, semiA, semiB, angle
    type(SourceParameters), intent(inout) :: source

    ! Local variables
    integer(int32) :: i, j
    real(single) :: x0, y0, theta, thetar
      
    real(double) :: f, fb, rawCounts, correctedCounts, countRate
    real(double) :: backgroundRate
    real(double) :: backgroundCounts = 0.0_double, correctedBackgroundCounts
    real(double) :: correction, pixArea, totalArea, coincLossArea
 
 
    integer :: ncount
          

    integer, parameter :: maxSize = 10000
    real(single) :: innerbckSemia, outerbcksemia
    real(single) :: innerbckSemib, outerbcksemib
    real(single) :: theBackground
    integer :: imin, imax, jmin, jmax


    coincLossArea = PI * coincLossRadius * coincLossRadius

    ccdImage => image
    ! Determine the corrected background count-rate per pixel using 
    if (angle .lt. 0.0) then
        theta = angle + 180.0
    else
        theta = angle
    end if

    thetar = theta * degreesToRadians


    nx = ubound(image, 1)

    ny = ubound(image, 2)

    xLeft = 0.0
    yBottom = 0.0
    xRight = real(nx)
    yTop = real(ny)
           
    il = 0
    iu = nx - 1
    jl = 0
    ju = ny - 1     


    !***************************************************************************
    ! Set up a background interpolant using the pixels in the background annulus
    !************************************************************************** 

  


    innerbcksemia = source%innerBckRadius
    outerbckSemia = source%outerBckRadius
    innerbcksemib = source%innerBckRadius / source%axialRatio
    outerbcksemib = source%outerBckRadius / source%axialRatio

    ncount = 0
    imin =  max(0, int(floor(source%xPos) - ceiling(outerbcksemiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + outerbcksemiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(outerbcksemiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + outerbcksemiA + 0.1) + 1) 
 



 

   
    theBackground = source%rawBackgroundCountRate
    rawCounts = 0.0_double
    correctedCounts = 0.0_double
    totalArea = 0.0_double
  
 


    imin =  max(0, int(floor(source%xPos) - ceiling(semiA)) - 1 )
    imax =  min(iu, int(ceiling(source%xPos) + semiA + 0.1) + 1) 
 
    jmin =  max(0, int(floor(source%yPos) - ceiling(semiA)) - 1 )
    jmax =  min(ju, int(ceiling(source%yPos) + semiA + 0.1) + 1) 
 

    meanExposure = 0.0_double
    numExposurePixels = 0

    do i = imin, imax
        x0 = real(i) + 0.5
     
        do j = jmin, jmax
       
                            
            y0 = real(j) + 0.5 ! 
            pixArea = PixelArea(x0 + 0.5, y0 + 0.5, source%xPos, source%yPos, semiA, semiB, theta)



            if(associated(exposureImage)) then
                meanExposure = meanExposure + dble(myExposureImage(i, j))
                numExposurePixels = numExposurePixels + 1
            end if 



            if (pixArea <= 0.0) cycle
            totalArea = totalArea + pixArea

            ! Compute the coincidence-loss correction for this pixel
            countRate= GetApertureCounts1(x0, y0, coincLossRadius) / source%exposure 
 

            !***************************************************************************
            ! Use the given background counts/pixel
            !**************************************************************************
            fb = theBackground 
          

            if (pixArea .lt. 1.0) then                  
                f =  PixelValue2(i, j) * pixArea            
                fb = fb * pixarea 
            else
                f =  PixelValue2(i, j)            
              
            end if
            backgroundCounts = backgroundCounts + fb
            rawCounts = rawCounts + f
            correctedCounts = correctedCounts + f * &
                        CoincidenceLossCorrection(countRate, sqrt(real(countRate)), myframeTime)
          
!image(i, j) = fb                       
        end do
    end do


!call SaveImage(image, "omsourceimage.fits", "test")








!write(*, *) "OmsourceExtendedSourceCounts2 semiA=", semia, " semiB=", semib, "angle=", angle, " total area=", totalarea
  
!write(*, *) "exposure=", exposure, " coinclossarea=", coinclossarea 


    backgroundRate = coincLossArea * backgroundCounts / (totalArea * source%exposure) ! Background count rate

!write(*, *) "background counts=", backgroundcounts, " backgroundrate=", backgroundrate


    ! The raw background count-rate
    source%rawBackgroundCountRate = backgroundCounts / source%exposure
      


    !******************************************************************
    ! The coincidence-loss correction for the background count-rate
    !******************************************************************
    correction = CoincidenceLossCorrection(backgroundRate, sqrt(real(backgroundRate)), myframeTime)

!write(*, *) "Coincidence-loss correction for background=", correction
!write(*, *) "background counts=", backgroundcounts, " total area=", totalarea, " coinclossarea=", coinclossarea
    
    ! The coincidence-loss corrected background count-rate-rate 
    source%corrBackGroundCountRate = source%rawBackgroundCountRate * correction ! Corrected background rate




    correctedBackgroundCounts = backgroundCounts * correction 

!write(*, *) "corrected total counts=", correctedcounts, " correctedbackgroundcounts=", correctedBackgroundCounts 


    !****************************************************
    ! The source coincidence-loss correction
    !****************************************************
    if (source%rawCountRate .gt. 0.0) then
        source%rightAscension =  correctedCounts / rawCounts
    end if

    ! The background coincidence-loss correction
    if (source%rawBackgroundCountRate .gt. 0.0) then
        source%declination =  correction
    end if



    !********************************************************
    ! The background subtracted raw count-rate
    !*******************************************************
    rawCounts = rawCounts - backgroundCounts  
    source%rawCountRate = rawCounts / source%exposure

    !*****************************************
    ! Recalculate the source significance
    !*****************************************   
    source%Significance = ComputeSourceSignificance(real(rawCounts), real(backgroundCounts))

 
    source%rawCountRateErr = ComputeSourceRateError(rawCounts, backGroundCounts, source%exposure)
               

    ! Corrected background-subtracted counts
    source%corrCountRate = (correctedCounts - correctedBackgroundCounts)  / source%exposure
    source%corrCountRateErr =  ComputeSourceRateError(correctedCounts, correctedBackGroundCounts, source%exposure) 

      
    source%psf1Correction = 0.0


    if (source%rawCountRate .gt. 0.0) then
        source%coincidenceLossCorr = source%corrCountRate / source%rawCountRate 
    else
       source%coincidenceLossCorr = 1.0
    end if
  
 
    nullify( ccdImage )
    

end subroutine extendedSourcePhotometry2      





!**************************************************************************************
!
! THE FOLLOWING IS FROM OMPHOTOM
!
!***************************************************************************************









!*********************************************
! Subroutine GetImageSize
!********************************************
subroutine GetImageSize(oswDataSet, windowX0,  windowY0,  windowDX, windowDY)
    type(DataSetT), intent(in) :: oswDataSet 
    integer(int32), intent(out) ::  windowX0,  windowY0,  windowDX, windowDY
    
    type(TableT) :: inTableH
    integer, parameter :: noColumns = 4
    type(ColumnT) :: inColumnH(noColumns)
    integer(int32) :: i
    integer(int32), pointer ::  windowX0P(:),  windowY0P(:),  windowDXP(:), windowDYP(:)
   
    

    if (hasAttribute(oswDataSet, "WINDOWX0") .and. &
        hasAttribute(oswDataSet, "WINDOWY0") .and. &
        hasAttribute(oswDataSet, "WINDOWDX") .and. &
        hasAttribute(oswDataSet, "WINDOWDY")) then

        windowx0 = int32Attribute(oswdataset, "WINDOWX0")
        windowy0 = int32Attribute(oswdataset, "WINDOWY0")
        windowdx = int32Attribute(oswdataset, "WINDOWDX")
        windowdY = int32Attribute(oswdataset, "WINDOWDY")
        return
    end if 

    !**************************************************
    ! Get a handle on the modes binary table extension.
    !**************************************************


    inTableH = table(oswDataSet, "MODES")

    !**************************************************
    ! Get handles and pointers on the modes binary table 
    ! extension columns and extract the window data.
    !**************************************************
    inColumnH(1) = column(inTableH, "WINDOWX0", read)
    windowX0P => int32Data(inColumnH(1))
    windowX0 = windowX0P(0)

    inColumnH(2) = column(inTableH, "WINDOWY0", read)
    windowY0P => int32Data(inColumnH(2))
    windowY0 = windowY0P(0)

    inColumnH(3) = column(inTableH, "WINDOWDX", read)
    windowDXP => int32Data(inColumnH(3))
    windowDX = windowDXP(0)

    inColumnH(4) = column(inTableH, "WINDOWDY", read)
    windowDYP => int32Data(inColumnH(4))
    windowDY = windowDYP(0)


    !************************************
    ! Release the handles
    !***********************************
    do i = 1, noColumns
        call release(inColumnH(i))
    end do
    call release(inTableH)

end subroutine GetImageSize


!***************************************************************
! Gets the background from the background image
!
!***************************************************************
subroutine getFastBackground( source  )

    type(SourceParameters), intent(inout) :: source
 
    real(single) :: xPos, yPos, mode, error

    !******************************************************************
    ! Get the full-frame image coordinates of the source
    !******************************************************************
    xPos = source%xpos
    yPos = source%ypos



    call ConvertCoordsTofullFrame( xPos, yPos, X0image, Y0image, imageBinX, imageBinY)


    il = 0
    jl = 0
    iu = real(nx - 1)
    ju = real(ny - 1)
    !*************************************************************************
    ! Convert to x, y image coordinates of background image
    !*************************************************************************           

    call ConvertCoordsFromfullFrame( xPos, yPos, X0Background, Y0Background, &
                                     backgroundBinX, backgroundBinY)
 

    call initialiseArrays


    ! Write out the source list to the oswDataSet
    write(errstr,'(a, a)') "Obtaining background from ", trim(stringParameter("backgroundimageset")) 
    call message(AppMsg, Verbose)


    !*************************************************************
    ! Get the background mode
    !*************************************************************
    mode = GetBackGroundMode4(  &
        mybackgroundImage,      &  
        xpos - 0.5,             &   ! The centre x-pixel coordinate of the aperture
        ypos - 0.5,             &   ! The centre y-pixel corrdinate of the aperture 
        source%innerBckRadius,  &   ! The aperture inner radius
        source%outerBckRadius,  &   ! The aperture outer radius
        error)      ! The error

    call deleteArrays

    source%rawBackgroundCountRate = mode

    write(errstr,'(a, f6.2)') "Computed background mode=", mode
    call message(AppMsg, Verbose)




end subroutine getFastBackground




!***************************************************************
! For the FAST-MODE if an image from the IMAGING chain is given
! to get the background, then need to make the image the
! same sze
!***************************************************************
subroutine getParameters
  
    
    type(DataSetT) :: dataSetH
    integer :: dx, dy 
    logical :: binBpe
     
    !***********************************************
    ! Get the window parameters for the image 
    !**********************************************
    dataSetH = dataSet( trim(stringParameter("imageset")), read)
    call GetImageSize( dataSetH, X0Image,  Y0Image,  dx, dy)
    call release(dataSetH)

    imageBinx = binx

    imageBiny = biny

    !*****************************************************
    ! Get the window parameters for the background image 
    !******************************************************



    dataSetH = dataSet( trim(stringParameter("backgroundimageset")), read)
    call GetImageSize( dataSetH, X0Background,  Y0Background,  dx, dy)


    ! compute binning
    binBpe = booleanAttribute(dataSetH, "BINBPE")
    backgroundBinx = 2 ** int32Attribute(dataSetH, "BINAX1")
    backgroundBiny = 2 ** int32Attribute(dataSetH, "BINAX2")


    if (binBpe) then
        backgroundBinx= backgroundBinx * 2
        backgroundBiny= backgroundBiny * 2
    end if
    
    call release(dataSetH)






end subroutine getParameters







!***************************************************************************
! For psf photometry
!***************************************************************************




    !*****************************************************************
    !
    !*****************************************************************
!    call initialisePSFPhotometry(myimage, myfilter, myapertureRadius, mycalRadius)

 



    








!*****************************************************************
!
!*****************************************************************
subroutine initialisePSFPhotometry()




    !**************************************************
    ! Create the psf image
    !*************************************************
    call CreatethePsfImage
    




end subroutine initialisePSFPhotometry





!*************************************************************************
!
!
!*************************************************************************
subroutine loadBackgroundImage

    type(dataSetT) :: dataSetH

    integer(int32) ::  windowDX, windowDY
    logical :: binBpe


    myBackgroundImageIsLoaded = .true. ! Record that we are using a background image

    dataSetH = dataSet( trim(stringParameter("backgroundimageset")), read)


    call createImageArray( dataSetH, myBackgroundImage)



    
    if( (ubound(myBackgroundImage, 1) .eq. ubound(myImage, 1)) .or. &
        (ubound(myBackgroundImage, 2) .eq. ubound(myImage, 2)) ) then
        fastBackground = .false.


        call release(dataSetH)
        return

    elseif ( trim(stringAttribute(myImageSet, "DATAMODE")) .eq."IMAGING") then

        call fatal("Invalid background image",  trim(stringParameter("backgroundimageset")))


    end if
    fastBackground = .true.



    !*********************************************************************
    ! Need some further image details
    !*********************************************************************


    call GetImageSize(myImageSet, x0Image,  y0Image,  windowDX, windowDY)
   

    !*********************************************************************
    ! Need to check the dimensions of the background image
    !*********************************************************************


    call GetImageSize(dataSetH, x0Background,  y0Background,  windowDX, windowDY)
   
 

    !****************************************************************************
    ! Get the binning factors
    !***************************************************************************
   
    binBpe = booleanAttribute(dataSetH, "BINBPE")
    backgroundBinx = 2 ** int32Attribute(dataSetH, "BINAX1")
    backgroundBiny = 2 ** int32Attribute(dataSetH, "BINAX2")

    if (binBpe) then
        backgroundBinx = backgroundBinx * 2
        backgroundBiny = backgroundBiny * 2
     
    end if
    
 

    call release( dataSetH )
        

end subroutine loadBackgroundImage












!*****************************************************
!
!
!*****************************************************
subroutine CreatethePsfImage
    real(double) :: maxPsf
   
    real(single) ::  counts
    integer :: oldNx, oldNy

    integer i

    real(single), pointer, dimension(:) :: x=>null(), y=>null()




      




    oldnx = nx
    oldny = ny

   
    psfnx = 12
    psfny = 12
       
     

    l0 = 14 / binx

    if(associated( myPsfImage )) then
        deallocate( myPsfImage )
        nullify( myPsfImage )
    end if



    ! Get the point-spread function image
  
    call CreatePSFImage21(psfNx, psfNy, mypsfImage, 1024.5_double, 1024.5_double, &
        ratio=0.5_double, bin=binX, res=1) 




!    write(*, *) "Counts=", counts, " area=", area, " total counts from image", sum(mypsfImage)


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  


    psfXcentre = real(psfnx) / 2.0 ! The x centre of the psf image
    psfYcentre = real(psfny) / 2.0 ! The y centre of the psf image




    !*******************************************************
    ! Normalize psf image to a maximum height of 1.0
    !*******************************************************

    maxPsf = maxval(myPsfImage) 
    mypsfImage = mypsfImage / maxPsf


    call GetCounts2( mypsfImage, psfXCentre, psfYCentre, real(12 / binx), counts)

    sumPsfImage = counts


    x => allocateSingle(1, psfnx, "x")
    y => allocateSingle(1, psfnx, "y")

    do i = 1, psfnx
        x(i) = real(i) - 0.5
        y(i) = real(i) - 0.5

    end do


    !*************************************************************************
    !
    !*************************************************************************
    call gridInterpolantConstruct(psfnx, psfny, x, y, myPsfImage, psfgridfit)

!x = x + 0.5
!y = y + 0.5
!do i = 1, psfnx 
!do j = 1, psfny 
!write(*, *) "pixel ", i, ", ", j, " x=", x(i), " y=", y(j), " image=", myPsfImage(i-1, j-1), &
!" interpolated=", gridInterpolate(x(i), y(j), psfgridfit), " bilinear interpolated=", &
!bilinearInterpolate(mypsfimage, x(i), y(j))
!end do
!end do

!    call gridInterpolantDestruct( psfgridFit )

    deallocate(x, y)
    nullify(x, y)
    nx = oldnx
    ny = oldny    
 end subroutine CreatethePsfImage

!*************************************************************************
! Least-squares fit to the PSF, not including the background
!
!
!************************************************************************

subroutine PSFPhotometry1(image, source, innerA, outerA)
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters), intent(inout) :: source
    real(single), intent(in) :: innerA, outerA 

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    real(single), pointer, dimension(:) :: xdata => null(), ydata => null(), weight => null()
  
    real(single) :: xpos, ypos, dx, dy

  
    real(single) :: gradient, gradientError
 
     
    real(single) :: mode, error


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    imin = max(0, i0 - l0)
    imax = min(nx, i0 + l0)

    jmin = max(0, j0 - l0)
    jmax = min(ny, j0 + l0)
   

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocateSingle(1, n, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
  


 
    if( .not. fastBackground) then
        call initialiseArrays
        ccdImage => image
        !*************************************************************
        !
        !*************************************************************
        mode = GetBackGroundMode3(  &
        source%xpos - 0.5,          &   ! The centre x-pixel coordinate of the aperture
        source%ypos - 0.5,          &   ! The centre y-pixel corrdinate of the aperture 
        innerA, &   ! The aperture inner radius
        outerA, &   ! The aperture outer radius
        error)      ! The error
        call deleteArrays
        nullify(ccDImage)
    else
        mode = source%rawBackgroundCountRate
    end if

    do i = imin, imax

        dx = real(i + 1) - source%xpos  ! NB (1, 1) corresponds to pixel 0, 0

        xpos = psfXcentre + dx 
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
        do j = jmin, jmax
          
           
 
            dy = real(j + 1) - source%yPos 
     
            ypos = psfYcentre + dy
  
            if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
            n = n + 1
           
            !********************************************************
            ! NB pixel 0, 0 has coordinates 0.5, 0.5
            !********************************************************
       

            xData(n) = PsfInterpolation(xPos, yPos)

            if (xData(n) .lt. 0.0) then
                n = max(0, n - 1)
                cycle
            end if

            ydata(n) =  image(i, j)


!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


           
            weight(n) = weightingFunction( ydata(n)) 
            

        end do
    end do

  
    !************************************************************************
    ! Least-squares fit with no constant term
    !************************************************************************   

    call LeastSquaresFit1(n, xdata, ydata, weight, gradient, gradientError)


    gradient = gradient * sumPsfImage

  

    source%apertureRadius = myApertureRadius
 
    source%rawCountRate = gradient / source%exposure
    source%rawBackgroundCountRate = mode * PI * real(144  / &
                  (binX * binX)) / source%exposure



    call PointSourceCoincLossCorr( source, myApertureRadius, frameTime )     



!  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
!" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)





    
end subroutine PSFPhotometry1

!*************************************************************************
!
! Least-squares to the PSF to include the background
!
!************************************************************************

subroutine PSFPhotometry2(image, source)
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters), intent(inout) :: source
   

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, imin, imax, jmin, jmax, n
    
  
    real(single) :: xpos, ypos, dx, dy

  
    real(single) :: gradient(1), gradientError(1), constant, constantError
    
   
 
    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    i0 = max(0, int(source%xpos - 0.5))
    j0 = max(0, int(source%ypos - 0.5))

    imin = max(0, i0 - l0)
    imax = min(nx, i0 + l0)

    jmin = max(0, j0 - l0)
    jmax = min(ny, j0 + l0)
   

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, 1, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
  

    do i = imin, imax

        dx = real(i + 1) - source%xpos

        xpos = psfXcentre + dx
  
        if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
        do j = jmin, jmax
          
          
 
            dy = real(j + 1) - source%ypos
     
            ypos = psfYcentre + dy 
  
            if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
            n = n + 1
           
            xData(n, 1) = PsfInterpolation(xPos, yPos)


            if (xData(n, 1) .lt. 0.0) then
                n = max(0, n - 1)
                cycle
            end if

            ydata(n) =  image(i, j)

        
     
!            write(*, *) "i=", i, " j=", j, " dx=", dx, " dy=", dy, " xpos=", xpos, " ypos=", ypos, &
!              " psfimage=", xdata(n), " image=", ydata(n)




!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************

            weight(n) = weightingFunction( ydata(n)) 

           

        end do
    end do





    !**********************************************************************************
    ! Least-squares fit with constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit2(n, xdata=xdata, ydata=ydata, &
            weight=weight, nterms=1_int8,              &
            gradient=gradient, gradienterror=gradientError, &
            constant=constant, constanterror=constantError)







!write(*, *) "PSFphotometry2 gradient=", gradient, " constant=", constant
    gradient = gradient * sumPsfImage 

!  write(*, *) "Original raw counts =", source%rawCountRate * source%exposure, & 
!" ratio (new/old)=",  (gradient / (source%rawCountRate * source%exposure) )
 

    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)




    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    source%rawCountRate = gradient(1) / source%exposure

    source%apertureRadius = myApertureRadius
    source%rawBackgroundCountRate = constant * PI * real(144  / &
                  (binX * binX)) / source%exposure



    call PointSourceCoincLossCorr( source, myApertureRadius, frameTime )     



    
end subroutine PSFPhotometry2




!*************************************************************************
!
! Least-squares to the PSF to include the background and an arbitrary
! number of linear constants
! ie y=constant+a1*x+a2*x, etc
!
!************************************************************************

subroutine PSFPhotometry3(nterms, image, source)
    integer(int8), intent(in) :: nTerms
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters) :: source(1 : nterms)
   

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, k, n
   
  
    real(single) :: xpos, ypos, dx, dy
    
  
   
 

    real(single), dimension(1 : nterms) :: gradient, gradientError

    real(single) :: constant, constantError


    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)



!write(*, *) "omsource_old.f90 line 4519 imin=", imin, " imax=", imax
!write(*, *) "omsource_old.f90 line 4519 jmin=", jmin, " jmax=", jmax


    do i = 1, nTerms

        i0 = max(0, int(source(i)%xpos - 0.5))
        j0 = max(0, int(source(i)%ypos - 0.5))

        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(nx, i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(ny, j0 + l0))
!write(*, *) i, ") source%xpos=", source(i)%xpos, " y=", source(i)%ypos


    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    


!write(*, *) "omsource_old.f90 line 4539 imin=", imin, " imax=", imax, &
!  " jmin=", jmin, " jmax=", jmax, " i0=", i0, " j0=", j0


    xdata => allocate2dSingle(1, n, 1, int(nterms, int32), "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************

           
            n = n + 1
            do k = 1, nTerms 
                dx = real(i + 1) - source(k)%xpos

                xpos = psfXcentre + dx
                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - source(k)%ypos
     
                ypos = psfYcentre + dy
 
             
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
             
           
           

 
                xData(n, k) = PsfInterpolation(xPos, yPos)
 !               write(*, *) "line 4581 n=", n, " k=", k, " xdata=", xData(n, k)

            end do
            if(all(xdata(n, 1: nterms) < 1e-10)) then
                n = n - 1
            else


                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(n) = image(i, j)
                
 !         write(*, *) "line 4594 n=", n, " k=", k, " ydata=", yData(n)



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(n) = weightingFunction( ydata(n)) 
                
                        
                
            end if



        end do
    end do





!write(*, *) "Line 4619 n=", n, " nterms=", nterms
    !**********************************************************************************
    ! Least-squares fit with constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit2(n, xdata=xdata, ydata=ydata,             &
            weight=weight, nterms=int(nterms, int8),               &
            maxiter=1_int8, nstd1=3.0,                             &
            gradient=gradient, gradienterror=gradientError,        &
            constant=constant, constanterror=constantError)




   
 
    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


!write(*, *) "constant=", constant
    
    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    do i = 1, nterms    
        source(i)%rawCountRate = gradient(i) * sumPsfImage / source(i)%exposure

        source(i)%apertureRadius = myApertureRadius
        source(i)%rawBackgroundCountRate = constant * PI * real(144  / &
                  (binX * binX)) / source(i)%exposure

!write(*, *) i, ") gradient=", gradient(i)

        call PointSourceCoincLossCorr( source(i), myApertureRadius, frameTime )     
    end do

    
end subroutine PSFPhotometry3










!*************************************************************************
!
! Least-squares to the PSF to exclude the background and an arbitrary
! number of linear constants
! ie y=a1*x+a2*x, etc
!
!************************************************************************

subroutine PSFPhotometry33(nterms, image, source, background)
    integer(int8), intent(in) :: nTerms
    real(single), dimension(:, :), pointer :: image


    type(SourceParameters) :: source(1 : nterms)
    real(single), intent(in) :: background

    integer :: nx, ny, psfNx, psfny
    integer :: i0, j0, i, j, k, n
   
  
    real(single) :: xpos, ypos, dx, dy
   
  
   
 

    real(single), dimension(1 : nterms) :: gradient, gradientError

   

    psfnx = size(mypsfimage, 1) 
    psfny = size(mypsfimage, 2)  

    nx = ubound(image, 1)
    ny = ubound(image, 2)


    meanBackground = background

    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)



    do i = 1, nTerms

        i0 = max(0, int(source(i)%xpos - 0.5))
        j0 = max(0, int(source(i)%ypos - 0.5))

        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(nx, i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(ny, j0 + l0))
    end do

    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    




    xdata => allocate2dSingle(1, n, 1, int(nterms, int32), "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")


    n = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************

           
            n = n + 1
            do k = 1, nTerms 
                dx = real(i + 1) - source(k)%xpos

                xpos = psfXcentre + dx
                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - source(k)%ypos
     
                ypos = psfYcentre + dy
 
             
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
             
           
                

                xData(n, k) = PsfInterpolation(xPos, yPos)


            end do
            if(all(xdata(n, 1: nterms) < 1e-10)) then
                n = n - 1
            else


                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(n) = max(0.0, image(i, j) - background) ! Remove the background
                



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(n) = weightingFunction( ydata(n)) 
           
                
            end if



        end do
    end do






    !**********************************************************************************
    ! Least-squares fit without constant (background) term
    !**********************************************************************************  
    call LeastSquaresFit3(n, xdata=xdata, ydata=ydata,             &
            weight=weight, nterms=int(nterms, int8),               &
            maxiter=1_int8, nstd1=3.0,                             &
            gradient=gradient, gradienterror=gradientError)




   
 
    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)



    
    !*******************************************************************
    ! THE NEW RAW COUNT RATES FOR A 12 radius aperture
    !*******************************************************************
    do i = 1, nterms    
        source(i)%rawCountRate = gradient(i) * sumPsfImage / source(i)%exposure

        source(i)%apertureRadius = myApertureRadius
        source(i)%rawBackgroundCountRate = meanBackground * PI * real(144  / &
                  (binX * binX)) / source(i)%exposure

        call PointSourceCoincLossCorr( source(i), myApertureRadius, frameTime )     
    end do

    
end subroutine PSFPhotometry33







!************************************************************************************
!
! Least-squares fit to PSF, including background, and including x and y
! for 1 or more sources
!
!***********************************************************************************
subroutine PSFPhotometry4(image, source, nterms)
    real(single), dimension(:, :), pointer :: image

    integer(int32), intent(in) :: nterms

    type(SourceParameters), intent(inout) :: source(1 : nterms)
  

    integer(int32) :: i0, j0
    integer(int32) :: n
  

 
  
   
    real(double) :: x(1 : nterms * 2)
    
!    logical(boolean) :: toFit(2)
    integer(int32) :: status 
    
    real(double) :: funcValue
    real(double), dimension(1 : nterms * 2) :: step, var
    real(double) :: simp, stopcr
    integer :: nop, maxFn, iquad, nloop

    integer :: i, j

    numberToFit = nterms


  


    nx = ubound(image, 1)
    ny = ubound(image, 2)






    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)


   
    do i = 1, nTerms
        j = (i - 1) * 2 + 1


        x(j) = source(i)%xpos 
        bl(j) = source(i)%xpos - real(maxShift / binx) 
        bu(j) = source(i)%xpos + real(maxShift / binx)

        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - real(maxShift / binx) 
        bu(j + 1) = source(i)%ypos + real(maxShift / binx)

  

        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))
        
        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(int(nx, int32), i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(int(ny, int32), j0 + l0))
    end do


    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, nterms, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")

 
!    toFit = .true.


  


!    call setNConstant( 100 )



    step = 0.02_double / dble(binx)





    maxFn = nterms * 5000
    stopCr = 0.0000001_double

    iquad = 1
    nop = nterms * 2
    nloop = 2 * nop

!****************************************************
! FOR TESTING PURPOSES
!****************************************************

!    simp = PSFFunction12(x) 



    simp = PSFFunction10(x) / 10.0




    do i = 1, 2 
        call omminim(x, step, nop, funcValue, maxFn, stopcr, nloop, iquad, &
                    simp, var, psfFunction10, status)  
        if(status .eq. 0) exit
    end do  

    if(status .ne. 0) then
        write(errstr, '(a, i5, a, e14.8)') "Status=", status, &
            " The fit may not be very good! - standard deviation of fit=", funcValue
        call message(AppMsg, Verbose)

    end if






    ! Write out the source list to the oswDataSet
    write(errstr,'(a, f7.2, a, f7.2, a, f7.2, a, f7.2 )') &
                "Old coordinates- x=", source(1)%xpos, " y=", source(1)%ypos

    call message(AppMsg, Verbose)
    
 
    write(errstr,'(a, f7.2, a, f7.2)') &
                "New coordinates- x=", x(1), " y=", x(2)
    
    call message(AppMsg, Verbose)



    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


    source(1)%apertureRadius = myApertureRadius
 
    source(1)%xpos = x(1)
    source(1)%ypos = x(2)









    call PSFPhotometry3(numberToFit, image, source)

    

end subroutine PSFPhotometry4



!************************************************************************************
!
! Least-squares fit to PSF, excluding background, but including x and y,
! for 1 or more sources
!
!***********************************************************************************
subroutine PSFPhotometry44(image, source, nterms, background)
    real(single), dimension(:, :), pointer :: image

    integer(int32), intent(in) :: nterms

    type(SourceParameters), intent(inout) :: source(1 : nterms)
    real(single), intent(in) :: background

    integer(int32) :: i0, j0
    integer(int32) :: n
  

 
  
   
    real(double) :: x(1 : nterms * 2)
    
!    logical(boolean) :: toFit(2)
    integer(int32) :: status 
    
    real(double) :: funcValue
    real(double), dimension(1 : nterms * 2) :: step, var
    real(double) :: simp, stopcr
    integer :: nop, maxFn, iquad, nloop

    integer :: i, j

    numberToFit = nterms

    meanBackground = background

  


    nx = ubound(image, 1)
    ny = ubound(image, 2)






    !******************************************************************
    ! Find out roughly how many pixels to be used in the fit
    !******************************************************************
    n = 0



    imin = ubound(image, 1)
    imax = lbound(image, 1)

    jmin = ubound(image, 2)
    jmax = lbound(image, 2)


   
    do i = 1, nTerms
        j = (i - 1) * 2 + 1


        x(j) = source(i)%xpos 
        bl(j) = source(i)%xpos - real(maxShift / binx) 
        bu(j) = source(i)%xpos + real(maxShift / binx)

        x(j + 1) =  source(i)%ypos 
        bl(j + 1) = source(i)%ypos - real(maxShift / binx) 
        bu(j + 1) = source(i)%ypos + real(maxShift / binx)

  

        i0 = max(0, int(bl(j) - 0.5))
        j0 = max(0, int(bl(j + 1) - 0.5))
        
        imin = min(imin, max(0, i0 - l0))
        imax = max(imax, min(int(nx, int32), i0 + l0))

        jmin = min(jmin, max(0, j0 - l0))
        jmax = max(jmax, min(int(ny, int32), j0 + l0))
    end do


    n = (imax - imin + 1) * (jmax - jmin + 1) ! The maximum number of pixels to be used

    xdata => allocate2dSingle(1, n, 1, nterms, "xdata")
    ydata => allocateSingle(1, n, "ydata")
    weight => allocateSingle(1, n, "weight")

 


    write(errstr, '(a, f7.2)') "Calling non-linear least-squares fitting routine- background fixed at ", background
    call message(AppMsg, Verbose)




    step = 0.2_double / dble(binx)
    maxFn = nterms * 5000
    stopCr = 0.0000001_double

    iquad = 1
    nop = nterms * 2
    nloop = 2 * nop





    simp = PSFFunction10(x) / 10.0




    do i = 1, 2
        call omminim(x, step, nop, funcValue, maxFn, stopcr, nloop, iquad, &
                    simp, var, psfFunction11, status)
        if(status .eq. 0) exit  
    end do  

    if(status .ne. 0) then
        write(errstr, '(a, i5, a, e14.8)') "Status=", status, &
         " The fit may not be too good! - Standard deviation of fit=", funcValue
        call message(AppMsg, Verbose)
    end if







    ! Write out the source list to the oswDataSet
    write(errstr,'(a, f7.2, a, f7.2, a, f7.2, a, f7.2 )') &
                "Old coordinates- x=", source(1)%xpos, " y=", source(1)%ypos

    call message(AppMsg, Verbose)
    
 
    write(errstr,'(a, f7.2, a, f7.2)') &
                "New coordinates- x=", x(1), " y=", x(2)
    
    call message(AppMsg, Verbose)



    deallocate(xdata, ydata, weight)
    nullify(xdata, ydata, weight)


    source(1)%apertureRadius = myApertureRadius
 
    source(1)%xpos = x(1)
    source(1)%ypos = x(2)









    call PSFPhotometry33(numberToFit, image, source, background)

    

end subroutine PSFPhotometry44






!***********************************************************
! 
! Fitting subroutine used by nonlinls
! 
!**********************************************************
function PSFFunction10(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
    real(single) :: constant, constantError
    integer :: i, j, k, l, ndata
    real(double) :: fit, residual

    gradientError = 0.0

    constantError = 0.0

    !****************************************************
    ! Check to make sure x values are within bounds
    !***************************************************
    do i = 1, numberToFit

        if(x(i) .lt. bl(i)) then
            functionValue = 1e20
            return 

        end if
 
        if(x(i) .gt. bu(i)) then
            functionValue = 1e20
            return 
        
        end if
    end do  

    nData = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy 
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                      

                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1

            else

                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(nData) = myimage(i, j)
                



!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************



                weight(nData) = weightingFunction( ydata(nData)) 
 
                                               
            end if

        end do
    end do



    !********************************************************************************
    ! Do the least-squares fit
    !*******************************************************************************  
    call LeastSquaresFit2(ndata, xdata(:,  1 : numberToFit), ydata=ydata,          &
            weight=weight, nterms=numberToFit, maxIter=1_int8, gradient=gradient,  &
            gradientError=gradientError, constant=constant,                        &
            constantError=constantError)



    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = dble(constant)

        do k = 1, numberToFit ! Sum for each source
            fit = fit + dble(gradient(k)) * dble(xdata(i, k))
        end do
  
        residual = fit - dble(ydata(i))

        functionValue = functionValue + residual * residual * dble(weight(i))

    end do 


   functionValue = sqrt(functionvalue /dble(ndata)) 

   

!write(*, *) "PSFFUNCTION10 x=", x, " functionvalue=", functionvalue


   
end function PSFFunction10




!***********************************************************
! 
! Fitting subroutine for use with omminim
! Least-squares fit to one or more sources, including 
! positions
! 
!**********************************************************
function PSFFunction11(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
  
    integer :: i, j, k, l, ndata
    real(double) :: fit, residual

    gradientError = 0.0

 

    !****************************************************
    ! Check to make sure x values are within bounds
    !***************************************************
    do i = 1, numberToFit

        if(x(i) .lt. bl(i)) then
            functionValue = 1e20
            return 

        end if
 
        if(x(i) .gt. bu(i)) then
            functionValue = 1e20
            return 
        
        end if
    end do  

    nData = 0
 
   
    do i = imin, imax

        do j = jmin, jmax



            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx 

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                       
    

                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1

            else

                !****************************************************
                ! Data point okay
                !*************************************************** 
                yData(nData) = myimage(i, j) - meanBackground
                


!**************************************************************************
! NB Set all weights to 1, because according ti Mighell 1989 MNRAS, 238, 807
! this is the best weighting procedure
!**************************************************************************


                weight(ndata) = weightingFunction( ydata(nData)) 
               

                                               
            end if

        end do
    end do



    !********************************************************************************
    ! Do the least-squares fit
    !*******************************************************************************  
    call LeastSquaresFit3(ndata, xdata(:,  1 : numberToFit), ydata=ydata,          &
            weight=weight, nterms=numberToFit, maxIter=1_int8, gradient=gradient,  &
            gradientError=gradientError)



    !****************************************************************
    ! Compute the weighted residuals
    !***************************************************************
    
    functionValue = 0.0_double

    do i = 1, ndata

        fit = 0.0_double

        do k = 1, numberToFit ! Sum for each source
            fit = fit + dble(gradient(k)) * dble(xdata(i, k))
        end do
  
        residual = fit - dble(ydata(i))

        functionValue = functionValue + residual * residual * dble(weight(i))

     

    end do 
   
    functionValue = sqrt(functionvalue /dble(ndata)) 


end function PSFFunction11










!***********************************************************
! 
! Fitting subroutine used by nonlinls
! 
!**********************************************************
function PSFFunction12(x) result(functionValue)
    real(double), intent(inout) :: x(:)
    real(double) :: functionValue
   

    real(single) :: xpos, ypos, dx, dy

    real(single), dimension(1 : numberToFit) :: gradient, gradientError
    real(single) :: constant, constantError
    integer :: i, j, k, l, ndata
    real(double) :: fit

    gradientError = 0.0

    constantError = 0.0

    nData = 0

    functionValue = 0.0_double



    constant = 10.0

    do i = 1, numberToFit

        gradient(i) = real(100 * i)

    end do
 
    do i = imin, imax

        do j = jmin, jmax


            !**********************************************************************
            ! Get the PSF counts for this pixel
            !**********************************************************************
           
            nData = nData + 1

            do k = 1, numberToFit
                l = (k  - 1) * 2 + 1
                dx = real(i + 1) - x(l)

                xpos = psfXcentre + dx

                if (xpos .lt. 0.0 .or. xpos .gt. real(psfnx)) cycle 
           
                dy = real(j + 1) - x(l + 1)
     
                ypos = psfYcentre + dy 
 
                if (ypos .lt. 0.0 .or. ypos .gt. real(psfny)) cycle 
                      

 
                xData(nData, k) = PsfInterpolation(xPos, yPos)


            end do

            if(all(xdata(nData, 1: numberToFit) < 1e-10)) then
                nData = nData - 1
            else


                fit = dble(constant)

               do k = 1, numberToFit ! Sum for each source
                   fit = fit + dble(gradient(k)) * dble(xdata(ndata, k))
               end do
  
 





                !****************************************************
                ! Data point okay
                !*************************************************** 
                myimage(i, j) = fit
                
                         
            
                                
            end if



        end do
    end do

end function PSFFunction12

!**************************************************
! Gets the 2-d image from the fits file
!**************************************************
subroutine createImageArray(dataSet, image, blockName)
    type(dataSetT), intent(in) :: dataSet
    real(single),   pointer    :: image(:, :)
    character(*), intent(in), optional :: blockName

    type(ArrayT)           :: handle
    integer(int8), pointer :: int8ImageP(:, :)     ! Integer8 array pointer.
    integer(int32),pointer :: int32ImageP(:, :)    ! Integer32 array pointer.
    integer(int16),pointer :: int16ImageP(:, :)    ! Integer16 array pointer.
    real(single),  pointer :: real32ImageP(:, :)   ! Real32 array pointer.
    real(double),  pointer :: real64ImageP(:, :)   ! Real64 array pointer.
    integer                :: sz(2)
    integer                :: status

    nullify(image)
    if( present(blockName) ) then
        handle = array(dataSet, trim(blockName), read)
    else
        handle = array(dataSet, "PRIMARY", read)
    end if
    sz = dimensions(handle)


    if (all(sz > 0)) then
        allocate( image(0 : sz(1) - 1, 0 : sz(2) - 1), stat = status)


        select case(arrayDataType(handle))
        case(integer8)
            int8ImageP => int8Array2Data(handle)
            image = real(int8ImageP)

        case(integer16)
            int16ImageP => int16Array2Data(handle)
            image = real(int16ImageP)

        case(integer32)
            int32ImageP => int32Array2Data(handle)
            image = real(int32ImageP)

        case(real32)
            real32ImageP => real32Array2Data(handle)
            image = real32ImageP

        case(real64)
            real64ImageP => real64Array2Data(handle)
            image = real64ImageP
        end select
    end if
 
    call release(handle)

end subroutine createImageArray
!*********************************************************
!
!
!*********************************************************
subroutine createQualityArray(dataSet,quality)
    type(dataSetT), intent(in) :: dataSet
    logical(boolean),  pointer    :: quality(:, :)

    type(ArrayT)                :: handle
    integer(int8), pointer :: int8ArrayP(:, :) => null()
    integer(int16), pointer :: int16ArrayP(:, :) => null()

    integer(int32)              :: sz(2)
    integer                     :: status

    nullify(quality)
    handle = array(dataSet, "QUALITY", read)

    sz = dimensions(handle)

    allocate( quality(0 : sz(1) - 1, 0 : sz(2) - 1), stat=status)


    if (all(sz > 0)) then
        


        select case(arrayDataType(handle))
            case(integer8)            
                int8ArrayP => int8Array2Data(handle)

                where (int8ArrayP .ne. 0)
                    quality = .true.
                elsewhere
                    quality = .false.
                end where
            case(integer16)
                int16ArrayP => int16Array2Data(handle)

                where (int16ArrayP .ne. 0)
                    quality = .true.
                elsewhere
                    quality = .false.
                end where
       
          end select
      
  
    end if

    call release(handle)


end subroutine createQualityArray
!*************************************************************
!
!************************************************************
real(single) function PsfInterpolation(xpos, ypos)
    real(single), intent(in) :: xPos, yPos




!    psfInterpolation = max(0.0, bilinearInterpolate(mypsfimage, xPos, yPos))
    psfInterpolation = max(0.0, gridInterpolate(xpos, ypos, psfgridfit))

end function psfInterpolation


!*********************************************************
!
!*********************************************************
real(single) function weightingFunction( data )
    real(single), intent(in) :: data
 

    

    if ( data .gt. 0.0 .and. weightingMethod .eq. 1) then
        

        weightingFunction = 1.0 / data



    else
        weightIngFunction  = 1.0
    end if



end function weightingFunction






!******************************************************************
!
!
!
!
!*****************************************************************
subroutine loadTheImage( dataSetH )
    type(DataSetT) :: dataSetH
  
  
   
    integer :: count = 0
    integer :: i, j
    integer(int16) :: nx, ny
    real(single), pointer :: counts(:) => null()

!write(*, *) "Load the image line 5800"

    if ( associated(myImage) ) then
        deallocate( myImage )
        nullify( myImage ) 
    end if

    call createImageArray( dataSetH, myImage)


   return
    !***********************************************************
    !
    ! Check to see if using an exposure image
    !
    !************************************************************

    if( hasBlock(dataSetH, "EXPOSURE") ) then 

!write(*, *) "Using exposure image"

         if ( associated(myExposureImage) ) then
             deallocate( myExposureImage )
             nullify( myExposureImage ) 
         end if

!write(*, *) "Line 5850"

         nx = ubound(myImage, 1)
         ny = ubound(myImage, 2)


        call createImageArray( dataSetH, myExposureImage, "EXPOSURE") 
    
!write(*, *) "Line 5857"


        counts => allocateSingle(1, int(nx, int32) * int(ny, int32), "counts")
           
        count = 0
   

        do j = 0, ny - 1
            do i = 0, nx - 1
                if( .not. isNullValue(myexposureImage(i, j) ) &
                    .and. myExposureImage(i, j) .gt. 0.0 ) then
                    count = count + 1
                    counts(count) = myexposureImage(i, j)
                                  
                end if
            end do
        end do

    
        exposureConstant = MedianValue( counts, count)
          

        do j = 0, ny - 1
            do i = 0, nx - 1
                if( .not. isNullValue(myImage(i, j) )) then
                    myImage(i, j) = myImage(i, j) * exposureConstant  

      
                end if
            end do
        end do
         
        call setExposureImage(myExposureImage, exposureConstant)
        deallocate(counts)
        nullify(counts)
    end if
      
!write(*, *) "exposure constant=", exposureConstant
!write(*, *) "leaving Load the image line 5800"
!write(*, *) "count=", count


end subroutine loadTheImage



!***********************************************************************
! This routine has been added in July 2008 to scale the count-rate errors
! when an exposure image is being used. 
! In such cases the exposure time of
! each source will have been set to 1.
! The variable exposureConstant will have been set to the mean exposure of
! the exposure image.
! The variable meanExposure will equal the mean exposure within the
! aperture area
! NB source%exposure will have been set equal to 1.0
!
!
!***********************************************************************
subroutine scaleRateErrors( source )
    type(SourceParameters), intent(inout) :: source

    real(double) :: constant, constant1
!    real(single) :: oldSignificance

!    write(*, *) "omphotometry - Scaling rate errors, etc, extensionflag=", source%extensionflag
!    write(*, *) "x=", source%xpos, " y=", source%ypos, " exposure=", source%exposure
!    write(*, *) "Input raw rate=", source%rawCountrate, " significance=", source%significance, &
!        " background=", source%rawBackgroundCountRate, " exposure=", source%exposure
!write(*, *) "coorate=", source%corrCountRate, " background=", source%corrCountRate
!write(*, *) "numexposurepixels=", numexposurepixels

    if (numExposurePixels .gt. 0) then

        

        !************************************************************************
        !
        ! Compute the mean exposure time within the photometry aperture
        !
        !***********************************************************************
        meanExposure = meanExposure / dble(numExposurePixels)



!        write(*, *) "mean exposure=", meanExposure, " numexposurepixels=", numExposurePixels
!        write(*, *) "exposureconstant=", exposureconstant       
     
!oldSignificance = source%significance



        !**********************************************************************
        !
        ! Need to divide all rates by the exposure constant
        !
        ! Rate * (mean exposure / exposure constant) gives the total counts
        ! (since the source%exposure is set to 1.0) and then divide by the
        ! mean exposure to get the rate 
        ! equals rate / exposure constant 
        !
        !**********************************************************************
 
 

        source%rawCountRate = source%rawCountRate / exposureConstant 
        source%rawBackgroundCountRate = source%rawBackgroundCountRate / exposureConstant 


        source%corrCountRate = source%corrCountRate / exposureConstant 
        source%corrBackgroundCountRate = source%corrBackgroundCountRate / exposureConstant 







        !******************************************************************
        !
        ! Scale the significance
        !
        !******************************************************************
        constant = meanExposure * meanExposure / exposureConstant   
        constant1 = sqrt( constant )  

!write(*, *) "constant=", constant, " constant1=", constant1
!write(*, *) "initial rawcountrateerr=", source%rawCountRateErr, " source%exposure=", source%exposure



      


        !******************************************************************
        !
        ! Scale the count-rate errors
        !
        !******************************************************************
        source%rawCountRateErr = source%rawCountRateErr / exposureConstant
        source%rawBackgroundCountRateErr = source%rawBackgroundCountRateErr / exposureConstant

     
        source%corrCountRateErr = source%corrCountRateErr / exposureConstant
        source%corrBackgroundCountRateErr = source%corrBackgroundCountRateErr / exposureConstant





!write(*, *) "Finalrawcountrateerr=", source%rawCountRateErr




        !***************************************************************************
        !
        ! Set the exposure-time equal to the mean exposure-time within the aperture
        !
        !***************************************************************************
        source%exposure = meanExposure
      
        source%significance = computeSourceSignificance(source%rawCountRate, &
              source%rawBackgroundCountRate, source%exposure)  


   
!        write(*, *) "Line 504 extension flag=" , Source%extensionFlag, &
!                  "   rawrate=", source%rawcountrate,      " significance=", source%significance, &
!                  " coorrate=", source%corrcountrate, " background=", source%corrbackgroundCountRate

!        write(*, *) "numexposurepixels=", numExposurePixels, " meanexposure=", meanExposure 

!write(*, *) "OLD SIG=", oldSignificance, " new=", source%significance

    end if


end subroutine scaleRateErrors



end module omsource_old










